import argparse
import json
import sys
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, cast

from ..dsl.by_yaml._internal.config_parsing.error_envelope import ErrorEnvelope, ScalimYamlValidationError
from ..dsl.by_yaml._internal.config_parsing.imports import ScalimYamlImportExpansionError, contains_import_syntax, expand_imports_inplace
from ..dsl.by_yaml._internal.config_parsing.jsonschema_issues import ScalimJsonSchemaCollectorError, collect_jsonschema_validation_issues
from ..dsl.by_yaml._internal.config_parsing.unknown_fields import UnknownFieldIssue, find_unknown_fields
from ..dsl.by_yaml._internal.config_parsing.validator import ConfigValidator
from ..dsl.by_yaml._internal.config_parsing.validators.issues import ValidationIssue
from ..dsl.by_yaml._internal.config_parsing.yaml_load import (
    YamlLocationIndex,
    envelope_from_validation_issue,
    error_loc_for_yaml_path,
    load_yaml_mapping_text,
)
from ..dsl.by_yaml.workflow_config import load_workflow_config_from_mapping
from ..dsl.by_yaml.workflow_paths import resolve_workflow_demand_path
from ..dsl.by_yaml.workflow_types import ScalimWorkflowConfigError
from ..exceptions import safe_error_message, safe_error_type
from ..vendor.compact.importlibx import import_module
from ..vendor.dataclassesx import dataclass, field
from ..vendor.yamlx import yaml
from . import yaml_dsl_lsp

try:
    jsonschema = import_module("jsonschema")
except ImportError:
    jsonschema = None  # type: ignore[assignment]
    _HAS_JSONSCHEMA = False  # pyright: ignore[reportConstantRedefinition]
else:
    _HAS_JSONSCHEMA = True


@dataclass
class ValidationPayload:
    mode: str
    """校验模式(例如 `validator`/`schema`)."""

    ok: bool
    """是否通过校验."""

    yaml_path: Optional[str] = None
    """可选:被校验的 `YAML` 文件路径."""

    schema_path: Optional[str] = None
    """可选:使用的 `JSON Schema` 文件路径."""

    errors: List[ErrorEnvelope] = field(default_factory=list)
    """错误列表."""

    warnings: List[ErrorEnvelope] = field(default_factory=list)
    """告警列表."""

    def as_dict(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "mode": self.mode,
            "ok": self.ok,
            "errors": [issue.as_dict() for issue in self.errors],
            "warnings": [issue.as_dict() for issue in self.warnings],
        }
        if self.yaml_path is not None:
            payload["yaml_path"] = self.yaml_path
        if self.schema_path is not None:
            payload["schema_path"] = self.schema_path
        return payload


@dataclass
class WorkflowValidationPayload:
    mode: str
    ok: bool
    workflow_yaml_path: str
    results: List[ValidationPayload] = field(default_factory=list)

    def as_dict(self) -> Dict[str, Any]:
        return {
            "mode": self.mode,
            "ok": self.ok,
            "workflow_yaml_path": self.workflow_yaml_path,
            "results": [item.as_dict() for item in self.results],
        }


LEGACY_FIELDS = {
    "relations_sql_like",
    "relations_graph",
    "foreign_key",
    "target",
    "from",
    "via",
    "column",
    "pk",
    "pk_transform",
    "derived",
    "key_transform",
    "primary",
}


def register(subparsers: Any) -> None:
    parser = subparsers.add_parser("yaml-dsl", help="YAML DSL utilities")
    _set_help_default(parser)
    yaml_subparsers = parser.add_subparsers(dest="yaml_dsl_command")

    validate_parser = yaml_subparsers.add_parser("validate", help="Validate YAML DSL via internal validator")
    _add_validate_args(validate_parser)
    validate_parser.set_defaults(func=_run_validate)

    schema_parser = yaml_subparsers.add_parser("schema", help="Schema helpers")
    _set_help_default(schema_parser)
    schema_subparsers = schema_parser.add_subparsers(dest="yaml_dsl_schema_command")

    schema_validate_parser = schema_subparsers.add_parser("validate", help="Validate YAML DSL via JSON Schema")
    _add_schema_validate_args(schema_validate_parser)
    schema_validate_parser.set_defaults(func=_run_schema_validate)

    schema_show_parser = schema_subparsers.add_parser("show", help="Print JSON Schema")
    schema_show_parser.set_defaults(func=_run_schema_show)

    schema_path_parser = schema_subparsers.add_parser("path", help="Print JSON Schema path")
    schema_path_parser.set_defaults(func=_run_schema_path)

    upsert_parser = yaml_subparsers.add_parser(
        "upsert-lsp-comment",
        help="Upsert YAML $schema modeline comment (JetBrains/RedHat)",
    )
    _add_upsert_lsp_comment_args(upsert_parser)
    upsert_parser.set_defaults(func=_run_upsert_lsp_comment)


def _set_help_default(parser: argparse.ArgumentParser) -> None:
    def _show_help(_args: argparse.Namespace) -> int:
        parser.print_help()
        return 2

    parser.set_defaults(func=_show_help)


def _add_validate_args(parser: argparse.ArgumentParser) -> None:
    _ = parser.add_argument("yaml_file", type=Path, help="YAML 文件路径")
    _ = parser.add_argument("--schema", "-s", type=Path, default=None, help="JSON Schema 文件路径")
    _ = parser.add_argument(
        "--type",
        dest="yaml_type",
        type=str,
        choices=["auto", "demand", "workflow"],
        default="auto",
        help="校验类型: auto/demand/workflow",
    )
    _ = parser.add_argument(
        "--path-alias",
        dest="path_aliases",
        type=str,
        action="append",
        default=[],
        help="仅 workflow validate: 需求路径别名,格式 <alias>=<path> (可重复)",
    )
    _ = parser.add_argument(
        "--allowed-yaml-root",
        dest="allowed_yaml_roots",
        type=Path,
        action="append",
        default=[],
        help="允许读取 YAML 的根目录(可重复);默认仅允许入口 YAML 所在目录",
    )
    _ = parser.add_argument("--json", action="store_true", help="输出 JSON 结果")
    _ = parser.add_argument("--verbose", "-v", action="store_true", help="显示详细错误信息")


def _add_schema_validate_args(parser: argparse.ArgumentParser) -> None:
    _ = parser.add_argument("yaml_file", type=Path, help="YAML 文件路径")
    _ = parser.add_argument("--schema", "-s", type=Path, default=None, help="JSON Schema 文件路径")
    _ = parser.add_argument("--json", action="store_true", help="输出 JSON 结果")
    _ = parser.add_argument("--verbose", "-v", action="store_true", help="显示详细错误信息")


def _add_upsert_lsp_comment_args(parser: argparse.ArgumentParser) -> None:
    _ = parser.add_argument(
        "paths",
        type=Path,
        nargs="+",
        help="一个或多个 YAML 文件路径",
    )
    _ = parser.add_argument(
        "--type",
        dest="schema_type",
        type=str,
        default=yaml_dsl_lsp.DEFAULT_SCHEMA_TYPE,
        help="Schema 类型(例如 demand/workflow)",
    )
    _ = parser.add_argument(
        "--schema-path",
        dest="schema_path",
        type=str,
        default=yaml_dsl_lsp.DEFAULT_SCHEMA_PATH,
        help="Schema base URL/dir 或完整 .json URL/path(默认使用内置 schema 目录)",
    )
    _ = parser.add_argument(
        "--comment-style",
        dest="comment_style",
        type=str,
        choices=list(yaml_dsl_lsp.COMMENT_STYLE_CHOICES),
        default=yaml_dsl_lsp.DEFAULT_COMMENT_STYLE,
        help="Schema modeline 风格: all/jetbrains/redhat",
    )


def _default_schema_path() -> Path:
    return Path(__file__).resolve().parents[1] / "dsl" / "by_yaml" / "schema" / "demand.gen.json"


def _resolve_schema_path(arg: Optional[Path]) -> Path:
    return arg.resolve() if arg is not None else _default_schema_path()


def _load_json_schema(schema_path: Path) -> Dict[str, Any]:
    with schema_path.open("r", encoding="utf-8") as handle:
        return json.load(handle)


def _find_legacy_field_errors(
    yaml_data: Dict[str, Any],
    *,
    source_path: str,
    locations: YamlLocationIndex,
) -> List[ErrorEnvelope]:
    errors: List[ErrorEnvelope] = []

    _collect_legacy_fields(errors, yaml_data, None, source_path=source_path, locations=locations)

    sources = yaml_data.get("sources", {})
    if isinstance(sources, dict):
        sources_dict = cast("Dict[str, Any]", sources)  # pragma: allow-cast yaml mapping typed narrowing for legacy field scan
        for source_id, source_data in sources_dict.items():
            if not isinstance(source_data, dict):
                continue
            source_data_dict = cast("Dict[str, Any]", source_data)  # pragma: allow-cast yaml mapping typed narrowing for legacy field scan
            _collect_legacy_fields(
                errors,
                source_data_dict,
                "sources.{}".format(source_id),
                source_path=source_path,
                locations=locations,
            )

    fields = yaml_data.get("fields", {})
    if isinstance(fields, dict):
        fields_dict = cast("Dict[str, Any]", fields)  # pragma: allow-cast yaml mapping typed narrowing for legacy field scan
        for field_id, field_data in fields_dict.items():
            if not isinstance(field_data, dict):
                continue
            field_data_dict = cast("Dict[str, Any]", field_data)  # pragma: allow-cast yaml mapping typed narrowing for legacy field scan
            _collect_legacy_fields(
                errors,
                field_data_dict,
                "fields.{}".format(field_id),
                source_path=source_path,
                locations=locations,
            )

    return errors


def _collect_legacy_fields(
    errors: List[ErrorEnvelope],
    data: Dict[str, Any],
    prefix: Optional[str],
    *,
    source_path: str,
    locations: YamlLocationIndex,
) -> None:
    for key in data:
        if key not in LEGACY_FIELDS:
            continue
        path = "{}.{}".format(prefix, key) if prefix else str(key)
        errors.append(
            ErrorEnvelope(
                code="yaml_legacy_field",
                message="Legacy field '{}' is not allowed".format(key),
                source_path=source_path,
                path=path,
                loc=error_loc_for_yaml_path(path, locations),
            )
        )


def _issues_to_rows(
    issues: Iterable[object],
    *,
    source_path: str,
    locations: YamlLocationIndex,
    default_code: str,
) -> List[ErrorEnvelope]:
    rows: List[ErrorEnvelope] = []
    for issue in issues:
        if isinstance(issue, ErrorEnvelope):
            rows.append(issue)
            continue
        if isinstance(issue, ValidationIssue):
            rows.append(
                envelope_from_validation_issue(
                    issue,
                    source_path=source_path,
                    locations=locations,
                    default_code=default_code,
                )
            )
            continue
        if isinstance(issue, UnknownFieldIssue):
            path = str(issue.path or "(root)")
            rows.append(
                ErrorEnvelope(
                    code=str(default_code),
                    message=str(issue.message),
                    source_path=source_path,
                    path=path,
                    loc=error_loc_for_yaml_path(path, locations),
                    suggestions=tuple(issue.suggestions),
                )
            )
            continue
        rows.append(
            ErrorEnvelope(
                code=str(default_code),
                message=str(issue),
                source_path=source_path,
                path="(root)",
                loc=error_loc_for_yaml_path("(root)", locations),
            )
        )
    return rows


def _find_removed_outputs_defaults_errors(
    yaml_data: Optional[Dict[str, Any]],
    *,
    source_path: str,
    locations: Optional[YamlLocationIndex],
    default_code: str,
) -> List[ErrorEnvelope]:
    if not isinstance(yaml_data, dict) or "outputs_defaults" not in yaml_data:
        return []
    path = "outputs_defaults"
    loc = None if locations is None else error_loc_for_yaml_path(path, locations)
    return [
        ErrorEnvelope(
            code=default_code,
            message=(
                "outputs_defaults was removed; move the book binding to each Excel output's `outputs[*].to.book`. "
                "Reuse it with YAML anchors (`_templates`) or `$import` if needed."
            ),
            source_path=source_path,
            path=path,
            loc=loc,
        )
    ]


def _find_demand_book_binding_errors(
    yaml_data: Optional[Dict[str, Any]],
    *,
    source_path: str,
    locations: Optional[YamlLocationIndex],
    available_book_ids: Optional[Set[str]],
    default_code: str,
) -> List[ErrorEnvelope]:
    if not isinstance(yaml_data, dict):
        return []

    available = tuple(sorted(available_book_ids or set()))
    errors: List[ErrorEnvelope] = []
    for ref_path, book_id in _extract_demand_outputs_book_refs(yaml_data):
        loc = None if locations is None else error_loc_for_yaml_path(ref_path, locations)
        if not book_id:
            errors.append(
                ErrorEnvelope(
                    code=default_code,
                    message=(
                        "Missing outputs to.book binding; set `outputs[*].to.book` explicitly. "
                        "Reuse the binding with YAML anchors (`_templates`) or `$import` if needed."
                    ),
                    source_path=source_path,
                    path=ref_path,
                    loc=loc,
                    suggestions=available,
                )
            )
            continue
        if available_book_ids is not None and book_id not in available_book_ids:
            errors.append(
                ErrorEnvelope(
                    code=default_code,
                    message=(
                        "Unknown book id referenced by outputs binding: {!r} "
                        "(declare resources.books.{} in demand or workflow.resources.books.{} in workflow)"
                    ).format(book_id, book_id, book_id),
                    source_path=source_path,
                    path=ref_path,
                    loc=loc,
                    suggestions=available,
                )
            )
    return errors


def _retry_enabled_missing_should_retry(retry_raw: object) -> bool:
    if not isinstance(retry_raw, dict):
        return False
    retry_dict = cast("Dict[str, Any]", retry_raw)  # pragma: allow-cast yaml retry mapping typed narrowing
    if retry_dict.get("enabled") is not True:
        return False
    should_retry_raw = retry_dict.get("should_retry")
    if not isinstance(should_retry_raw, str):
        return True
    return not bool(should_retry_raw.strip())


def _find_retry_enabled_missing_should_retry_errors(
    yaml_data: Optional[Dict[str, Any]],
    *,
    source_path: str,
    locations: Optional[YamlLocationIndex],
    default_code: str,
) -> List[ErrorEnvelope]:
    if not isinstance(yaml_data, dict):
        return []

    def _add_error(path: str) -> ErrorEnvelope:
        loc = None if locations is None else error_loc_for_yaml_path(path, locations)
        return ErrorEnvelope(
            code=default_code,
            message=(
                "retry.enabled=true requires non-empty should_retry "
                "(provide it in YAML, or rely on runtime driver injection to supply it during compile/run)"
            ),
            source_path=source_path,
            path=path,
            loc=loc,
        )

    errors: List[ErrorEnvelope] = []
    if _retry_enabled_missing_should_retry(yaml_data.get("retry")):
        errors.append(_add_error("retry.should_retry"))

    main_source_raw = yaml_data.get("main_source")
    if isinstance(main_source_raw, dict):
        main_source_dict = cast("Dict[str, Any]", main_source_raw)  # pragma: allow-cast yaml main_source mapping typed narrowing
        if _retry_enabled_missing_should_retry(main_source_dict.get("retry")):
            errors.append(_add_error("main_source.retry.should_retry"))

    sources_raw = yaml_data.get("sources")
    if isinstance(sources_raw, dict):
        for source_id, source_cfg_raw in cast("Dict[str, Any]", sources_raw).items():  # pragma: allow-cast yaml sources typed narrowing
            if not isinstance(source_cfg_raw, dict):
                continue
            source_cfg_dict = cast("Dict[str, Any]", source_cfg_raw)  # pragma: allow-cast yaml source config mapping typed narrowing
            if _retry_enabled_missing_should_retry(source_cfg_dict.get("retry")):
                errors.append(_add_error("sources.{}.retry.should_retry".format(source_id)))

    return errors


def _write_line(text: str) -> None:
    _ = sys.stdout.write(text + "\n")


def _write_line_stderr(text: str) -> None:
    _ = sys.stderr.write(text + "\n")


def _write_raw(text: str) -> None:
    _ = sys.stdout.write(text)


def _display_issue_path(path: str) -> str:
    return path or "(root)"


def _format_issue_location(yaml_path: Path, issue: ErrorEnvelope) -> str:
    if issue.line is None:
        return str(yaml_path)
    if issue.column is not None:
        return "{}:{}:{}".format(yaml_path, issue.line, issue.column)
    return "{}:{}".format(yaml_path, issue.line)


def _emit_source_snippet(issue: ErrorEnvelope, source_lines: Optional[List[str]], *, verbose: bool) -> None:
    if source_lines is None or issue.line is None:
        return
    total_lines = len(source_lines)
    if total_lines == 0:
        return
    line_no = max(1, min(issue.line, total_lines))
    if verbose:
        start = max(1, line_no - 1)
        end = min(total_lines, line_no + 1)
    else:
        start = line_no
        end = line_no

    _write_line("  |")
    for current in range(start, end + 1):
        text = source_lines[current - 1].rstrip("\n")
        _write_line("{:>4} | {}".format(current, text))
        if current == line_no:
            column = issue.column or 1
            pointer = " " * max(column - 1, 0)
            _write_line("  | {}^".format(pointer))


def _print_linter_result(
    yaml_path: Path,
    *,
    errors: List[ErrorEnvelope],
    warnings: List[ErrorEnvelope],
    verbose: bool,
    source_lines: Optional[List[str]],
) -> None:
    if not errors and not warnings:
        _write_line("OK {}".format(yaml_path.name))
        return

    for issue in errors:
        path_suffix = _display_issue_path(issue.path)
        message = issue.message
        if path_suffix != "(root)":
            message = "{} [{}]".format(message, path_suffix)
        _write_line("ERROR {} --> {}".format(message, _format_issue_location(yaml_path, issue)))
        _emit_source_snippet(issue, source_lines, verbose=verbose)
        if issue.suggestions:
            _write_line("help: {}".format(", ".join(issue.suggestions)))
        _write_line("")

    for issue in warnings:
        path_suffix = _display_issue_path(issue.path)
        message = issue.message
        if path_suffix != "(root)":
            message = "{} [{}]".format(message, path_suffix)
        _write_line("WARN {} --> {}".format(message, _format_issue_location(yaml_path, issue)))
        _emit_source_snippet(issue, source_lines, verbose=verbose)
        if issue.suggestions:
            _write_line("help: {}".format(", ".join(issue.suggestions)))
        _write_line("")

    summary_parts: List[str] = []
    if errors:
        summary_parts.append("{} error{}".format(len(errors), "" if len(errors) == 1 else "s"))
    if warnings:
        summary_parts.append("{} warning{}".format(len(warnings), "" if len(warnings) == 1 else "s"))
    summary = ", ".join(summary_parts) if summary_parts else "no issues"
    _write_line("Found {}.".format(summary))


def _render_result(
    yaml_path: Path,
    *,
    errors: List[ErrorEnvelope],
    warnings: List[ErrorEnvelope],
    verbose: bool,
    source_lines: Optional[List[str]],
) -> None:
    _print_linter_result(
        yaml_path,
        errors=errors,
        warnings=warnings,
        verbose=verbose,
        source_lines=source_lines,
    )


def _emit_error(
    message: str,
    *,
    json_output: bool,
    yaml_path: Optional[Path] = None,
    schema_path: Optional[Path] = None,
    mode: Optional[str] = None,
) -> None:
    if json_output:
        source_path = str(yaml_path) if yaml_path is not None else "(unknown)"
        payload = ValidationPayload(
            mode=mode or "error",
            ok=False,
            yaml_path=str(yaml_path) if yaml_path is not None else None,
            schema_path=str(schema_path) if schema_path is not None else None,
            errors=[
                ErrorEnvelope(
                    code="cli_error",
                    message=message,
                    source_path=source_path,
                    path="(root)",
                    loc=None,
                )
            ],
        )
        _write_line(json.dumps(payload.as_dict(), ensure_ascii=False))
        return
    _write_line_stderr("错误: {}".format(message))


def _infer_yaml_type(yaml_text: str) -> str:
    try:
        yaml_data = yaml.safe_load(yaml_text)
    except Exception:  # noqa: BLE001
        return "demand"
    if isinstance(yaml_data, dict) and "workflow" in yaml_data:
        return "workflow"
    return "demand"


def _parse_path_aliases(raw_values: Iterable[str]) -> Tuple[Optional[Dict[str, str]], Optional[str]]:
    output: Dict[str, str] = {}
    for raw in raw_values:
        item = str(raw or "").strip()
        if not item:
            continue
        if "=" not in item:
            return None, "Invalid --path-alias value: {!r} (expected <alias>=<path>)".format(item)
        alias, base = item.split("=", 1)
        alias = str(alias or "").strip()
        base = str(base or "").strip()
        if not alias:
            return None, "Invalid --path-alias value: {!r} (alias must be non-empty)".format(item)
        if not base:
            return None, "Invalid --path-alias value: {!r} (path must be non-empty)".format(item)
        output[alias] = base
    return output, None


def _extract_demand_book_ids(yaml_data: Optional[Dict[str, Any]]) -> Set[str]:
    if not isinstance(yaml_data, dict):
        return set()
    resources_obj = yaml_data.get("resources")
    if not isinstance(resources_obj, dict):
        return set()
    resources = cast("Dict[str, Any]", resources_obj)  # pragma: allow-cast yaml mapping typed narrowing
    books_obj = resources.get("books")
    if not isinstance(books_obj, dict):
        return set()
    books = cast("Dict[str, Any]", books_obj)  # pragma: allow-cast yaml mapping typed narrowing
    out: Set[str] = set()
    for raw_book_id in books:
        if not isinstance(raw_book_id, str):
            continue
        bid = str(raw_book_id or "").strip()
        if bid:
            out.add(bid)
    return out


def _extract_output_book_ref(item: Dict[str, Any], *, idx: int) -> Tuple[str, str]:
    to_raw_obj = item.get("to")
    if isinstance(to_raw_obj, dict):
        to_raw = cast("Dict[str, Any]", to_raw_obj)  # pragma: allow-cast yaml mapping typed narrowing
        book_raw = to_raw.get("book")
        if isinstance(book_raw, str) and book_raw.strip():
            return "outputs.{}.to.book".format(int(idx)), str(book_raw).strip()
    return "outputs.{}.to.book".format(int(idx)), ""


def _extract_demand_outputs_book_refs(yaml_data: Optional[Dict[str, Any]]) -> List[Tuple[str, str]]:
    """提取需求 `outputs` 中绑定到 `books` 的引用信息.

    返回 `(ref_path, book_id)` 列表:
    - 对于包含 `container` 的 `outputs[*]`(`CSV` 文件输出),不返回任何项.
    - 若某个输出缺失有效 `book_id`,仍返回其 `ref_path`,但 `book_id` 为空字符串.
    """
    if not isinstance(yaml_data, dict):
        return []
    outputs_raw = yaml_data.get("outputs")
    if not isinstance(outputs_raw, list):
        return []
    outputs_list = cast("List[Any]", outputs_raw)  # pragma: allow-cast yaml outputs typed narrowing

    out: List[Tuple[str, str]] = []
    for idx, item in enumerate(outputs_list):
        if not isinstance(item, dict):
            continue
        item_dict = cast("Dict[str, Any]", item)  # pragma: allow-cast yaml mapping typed narrowing
        if item_dict.get("container") is not None:
            continue
        out.append(_extract_output_book_ref(item_dict, idx=int(idx)))

    return out


def _validate_demand_yaml_text(
    yaml_text: str,
    *,
    yaml_path: Path,
    schema_path: Path,
    validator: Optional[ConfigValidator] = None,
    allowed_yaml_roots: Optional[Sequence[Path]] = None,
    available_book_ids: Optional[Set[str]] = None,
) -> Tuple[ValidationPayload, Optional[Dict[str, Any]], Optional[List[str]]]:
    source_lines: List[str] = yaml_text.splitlines()
    try:
        yaml_data_dict, locations, _lines = load_yaml_mapping_text(
            yaml_text,
            source_path=str(yaml_path),
            detect_duplicate_keys=True,
        )
    except ScalimYamlValidationError as exc:
        return (
            ValidationPayload(
                mode="validate",
                ok=False,
                yaml_path=str(yaml_path),
                schema_path=str(schema_path),
                errors=list(exc.errors),
                warnings=list(exc.warnings),
            ),
            None,
            source_lines,
        )

    try:
        if contains_import_syntax(yaml_data_dict):
            _ = expand_imports_inplace(yaml_data_dict, yaml_path=yaml_path, allowed_yaml_roots=allowed_yaml_roots)
    except ScalimYamlImportExpansionError as exc:
        logical_path = str(exc.logical_path or "(root)")
        errors = [
            ErrorEnvelope(
                code="yaml_import_expansion_error",
                message=str(exc),
                source_path=str(yaml_path),
                path=logical_path,
                loc=error_loc_for_yaml_path(logical_path, locations),
            )
        ]
        return (
            ValidationPayload(
                mode="validate",
                ok=False,
                yaml_path=str(yaml_path),
                schema_path=str(schema_path),
                errors=errors,
                warnings=[],
            ),
            None,
            source_lines,
        )

    demand_validator = validator or ConfigValidator(schema_path=str(schema_path))
    report = demand_validator.validate_report(
        yaml_data_dict,
        strict_unknown_fields=True,
        enable_jsonschema_validation=True,
    )
    errors = [
        envelope_from_validation_issue(
            issue,
            source_path=str(yaml_path),
            locations=locations,
            default_code="yaml_validate_error",
        )
        for issue in report.errors()
    ]
    warnings = [
        envelope_from_validation_issue(
            issue,
            source_path=str(yaml_path),
            locations=locations,
            default_code="yaml_validate_warning",
        )
        for issue in report.warnings()
    ]
    demand_book_ids = _extract_demand_book_ids(yaml_data_dict)
    effective_book_ids = set(demand_book_ids)
    if available_book_ids is not None:
        effective_book_ids.update(available_book_ids)
    errors.extend(
        _find_removed_outputs_defaults_errors(
            yaml_data_dict,
            source_path=str(yaml_path),
            locations=locations,
            default_code="yaml_validate_error",
        )
    )
    errors.extend(
        _find_demand_book_binding_errors(
            yaml_data_dict,
            source_path=str(yaml_path),
            locations=locations,
            available_book_ids=effective_book_ids,
            default_code="yaml_validate_error",
        )
    )
    errors.extend(
        _find_retry_enabled_missing_should_retry_errors(
            yaml_data_dict,
            source_path=str(yaml_path),
            locations=locations,
            default_code="yaml_validate_error",
        )
    )

    ok = not errors
    return (
        ValidationPayload(
            mode="validate",
            ok=ok,
            yaml_path=str(yaml_path),
            schema_path=str(schema_path),
            errors=errors,
            warnings=warnings,
        ),
        yaml_data_dict,
        source_lines,
    )


def _validate_demand_yaml_path(
    yaml_path: Path,
    *,
    schema_path: Path,
    validator: Optional[ConfigValidator] = None,
    allowed_yaml_roots: Optional[Sequence[Path]] = None,
    available_book_ids: Optional[Set[str]] = None,
) -> Tuple[ValidationPayload, Optional[Dict[str, Any]], Optional[List[str]]]:
    if not yaml_path.exists():
        errors = [
            ErrorEnvelope(
                code="yaml_file_not_found",
                message="YAML 文件不存在: {}".format(yaml_path),
                source_path=str(yaml_path),
                path="(file)",
                loc=None,
            )
        ]
        return (
            ValidationPayload(
                mode="validate",
                ok=False,
                yaml_path=str(yaml_path),
                schema_path=str(schema_path),
                errors=errors,
                warnings=[],
            ),
            None,
            None,
        )
    try:
        yaml_text = yaml_path.read_text(encoding="utf-8")
    except Exception:  # noqa: BLE001
        errors = [
            ErrorEnvelope(
                code="yaml_file_read_error",
                message="YAML 文件读取失败: {}".format(yaml_path),
                source_path=str(yaml_path),
                path="(file)",
                loc=None,
            )
        ]
        return (
            ValidationPayload(
                mode="validate",
                ok=False,
                yaml_path=str(yaml_path),
                schema_path=str(schema_path),
                errors=errors,
                warnings=[],
            ),
            None,
            None,
        )
    return _validate_demand_yaml_text(
        yaml_text,
        yaml_path=yaml_path,
        schema_path=schema_path,
        validator=validator,
        allowed_yaml_roots=allowed_yaml_roots,
        available_book_ids=available_book_ids,
    )


def _run_validate(args: argparse.Namespace) -> int:  # noqa: C901, PLR0912, PLR0915
    yaml_path = args.yaml_file.resolve()
    schema_path = _resolve_schema_path(args.schema)
    args_dict = vars(args)
    yaml_type = str(args_dict.get("yaml_type", "auto") or "auto").strip()
    if yaml_type == "auto":
        try:
            yaml_text = yaml_path.read_text(encoding="utf-8")
        except Exception:  # noqa: BLE001
            yaml_text = ""
        yaml_type = _infer_yaml_type(yaml_text)
    raw_aliases = list(args_dict.get("path_aliases", []) or [])
    path_aliases, alias_error = _parse_path_aliases(raw_aliases)
    if alias_error is not None:
        _emit_error(alias_error, json_output=bool(args.json), yaml_path=yaml_path, schema_path=schema_path, mode="validate")
        return 1
    allowed_yaml_roots = list(args_dict.get("allowed_yaml_roots", []) or [])

    if not schema_path.exists():
        _emit_error(
            "Schema 文件不存在: {}".format(schema_path),
            json_output=args.json,
            yaml_path=yaml_path,
            schema_path=schema_path,
            mode="validate",
        )
        return 1

    if yaml_type == "workflow":
        try:
            workflow_text = yaml_path.read_text(encoding="utf-8")
        except Exception:  # noqa: BLE001
            _emit_error(
                "YAML 文件读取失败: {}".format(yaml_path),
                json_output=args.json,
                yaml_path=yaml_path,
                schema_path=schema_path,
                mode="workflow-validate",
            )
            return 1

        workflow_source_lines: List[str] = workflow_text.splitlines()
        workflow_locations: YamlLocationIndex = {}
        workflow_errors: List[ErrorEnvelope] = []
        workflow_warnings: List[ErrorEnvelope] = []

        wf_config = None
        try:
            root, workflow_locations, _lines = load_yaml_mapping_text(
                workflow_text,
                source_path=str(yaml_path),
                detect_duplicate_keys=True,
            )
        except ScalimYamlValidationError as exc:
            workflow_errors.extend(list(exc.errors))
            workflow_warnings.extend(list(exc.warnings))
        else:
            try:
                wf_config = load_workflow_config_from_mapping(root)
            except ScalimWorkflowConfigError as exc:
                path = str(exc.path or "(root)")
                workflow_errors.append(
                    ErrorEnvelope(
                        code="workflow_validate_error",
                        message=str(exc),
                        source_path=str(yaml_path),
                        path=path,
                        loc=error_loc_for_yaml_path(path, workflow_locations),
                    )
                )
            except Exception as exc:  # noqa: BLE001
                safe_type = safe_error_type(exc)
                safe_msg = safe_error_message(exc) or ""
                workflow_errors.append(
                    ErrorEnvelope(
                        code="workflow_validate_error",
                        message="Unexpected error: {}: {}".format(safe_type, safe_msg),
                        source_path=str(yaml_path),
                        path="(root)",
                        loc=error_loc_for_yaml_path("(root)", workflow_locations),
                    )
                )

        demand_results: List[ValidationPayload] = []
        demand_source_lines: List[Optional[List[str]]] = []
        demand_yaml_data: List[Optional[Dict[str, Any]]] = []

        demand_validator = ConfigValidator(schema_path=str(schema_path))
        if wf_config is not None:
            workflow_book_ids = set((wf_config.resources.books or {}).keys())
            for run_idx, run in enumerate(wf_config.runs):
                demand_path: Optional[Path] = None
                try:
                    demand_path = resolve_workflow_demand_path(
                        str(run.demand),
                        workflow_yaml_path=str(yaml_path),
                        path_aliases=path_aliases,
                        run_id=str(run.id),
                        allowed_yaml_roots=allowed_yaml_roots,
                    )
                except ScalimWorkflowConfigError as exc:
                    wf_path = "workflow.runs.{}.demand".format(int(run_idx))
                    workflow_errors.append(
                        ErrorEnvelope(
                            code="workflow_validate_error",
                            message=str(exc),
                            source_path=str(yaml_path),
                            path=wf_path,
                            loc=error_loc_for_yaml_path(wf_path, workflow_locations),
                        )
                    )
                    demand_results.append(
                        ValidationPayload(
                            mode="validate",
                            ok=False,
                            yaml_path=str(run.demand),
                            schema_path=str(schema_path),
                            errors=[
                                ErrorEnvelope(
                                    code="demand_path_resolve_failed",
                                    message="Demand path resolve failed: {}".format(str(exc)),
                                    source_path=str(run.demand),
                                    path="(file)",
                                    loc=None,
                                )
                            ],
                            warnings=[],
                        )
                    )
                    demand_source_lines.append(None)
                    demand_yaml_data.append(None)
                    continue

                if not demand_path.exists():
                    workflow_errors.append(
                        ErrorEnvelope(
                            code="demand_file_not_found",
                            message="Demand YAML 文件不存在: {}".format(demand_path),
                            source_path=str(yaml_path),
                            path="workflow.runs.{}.demand".format(int(run_idx)),
                            loc=error_loc_for_yaml_path(
                                "workflow.runs.{}.demand".format(int(run_idx)),
                                workflow_locations,
                            ),
                        )
                    )
                    payload, demand_data, lines = _validate_demand_yaml_path(
                        demand_path,
                        schema_path=schema_path,
                        validator=demand_validator,
                        allowed_yaml_roots=allowed_yaml_roots,
                        available_book_ids=workflow_book_ids,
                    )
                    demand_results.append(payload)
                    demand_source_lines.append(lines)
                    demand_yaml_data.append(demand_data)
                    continue

                payload, demand_data, lines = _validate_demand_yaml_path(
                    demand_path,
                    schema_path=schema_path,
                    validator=demand_validator,
                    allowed_yaml_roots=allowed_yaml_roots,
                    available_book_ids=workflow_book_ids,
                )
                demand_results.append(payload)
                demand_source_lines.append(lines)
                demand_yaml_data.append(demand_data)
        workflow_ok = not workflow_errors

        workflow_payload = ValidationPayload(
            mode="workflow-validate",
            ok=workflow_ok,
            yaml_path=str(yaml_path),
            errors=workflow_errors,
            warnings=workflow_warnings,
        )

        results: List[ValidationPayload] = [workflow_payload]
        results.extend(demand_results)

        ok = workflow_ok and all(item.ok for item in demand_results)
        payload = WorkflowValidationPayload(
            mode="workflow-validate",
            ok=ok,
            workflow_yaml_path=str(yaml_path),
            results=results,
        )

        if args.json:
            _write_line(json.dumps(payload.as_dict(), ensure_ascii=False))
        else:
            _render_result(
                yaml_path,
                errors=workflow_payload.errors,
                warnings=workflow_payload.warnings,
                verbose=args.verbose,
                source_lines=workflow_source_lines,
            )
            for idx, item in enumerate(demand_results):
                demand_path_str = item.yaml_path or ""
                demand_path = Path(demand_path_str) if demand_path_str else Path("demand.yaml")
                _render_result(
                    demand_path,
                    errors=item.errors,
                    warnings=item.warnings,
                    verbose=args.verbose,
                    source_lines=demand_source_lines[idx] if idx < len(demand_source_lines) else None,
                )

        return 0 if ok else 1

    try:
        yaml_text = yaml_path.read_text(encoding="utf-8")
    except Exception:  # noqa: BLE001
        _emit_error(
            "YAML 文件读取失败: {}".format(yaml_path),
            json_output=args.json,
            yaml_path=yaml_path,
            schema_path=schema_path,
            mode="validate",
        )
        return 1

    payload, _yaml_data, source_lines = _validate_demand_yaml_text(
        yaml_text,
        yaml_path=yaml_path,
        schema_path=schema_path,
        allowed_yaml_roots=allowed_yaml_roots,
    )

    if args.json:
        _write_line(json.dumps(payload.as_dict(), ensure_ascii=False))
    else:
        _render_result(
            yaml_path,
            errors=payload.errors,
            warnings=payload.warnings,
            verbose=args.verbose,
            source_lines=source_lines,
        )

    return 0 if payload.ok else 1


def _run_schema_validate(args: argparse.Namespace) -> int:
    schema_path = _resolve_schema_path(args.schema)
    yaml_path = args.yaml_file.resolve()
    schema, exit_code = _load_schema_or_error(schema_path, yaml_path=yaml_path, args=args)
    if exit_code != 0 or schema is None:
        return exit_code

    try:
        yaml_text = yaml_path.read_text(encoding="utf-8")
    except Exception:  # noqa: BLE001
        _emit_error(
            "YAML 文件读取失败: {}".format(yaml_path),
            json_output=args.json,
            yaml_path=yaml_path,
            schema_path=schema_path,
            mode="schema-validate",
        )
        return 1
    source_lines: Optional[List[str]] = yaml_text.splitlines()

    try:
        yaml_data_dict, locations, _lines = load_yaml_mapping_text(
            yaml_text,
            source_path=str(yaml_path),
            detect_duplicate_keys=True,
        )
    except ScalimYamlValidationError as exc:
        return _emit_schema_result(
            yaml_path,
            schema_path,
            list(exc.errors),
            list(exc.warnings),
            args,
            ok=False,
            source_lines=source_lines,
        )

    jsonschema_module = _get_jsonschema_module(args, yaml_path=yaml_path, schema_path=schema_path)
    if jsonschema_module is None:
        return 1

    try:
        if contains_import_syntax(yaml_data_dict):
            _ = expand_imports_inplace(yaml_data_dict, yaml_path=yaml_path)
    except ScalimYamlImportExpansionError as exc:
        logical_path = str(exc.logical_path or "(root)")
        errors = [
            ErrorEnvelope(
                code="yaml_import_expansion_error",
                message=str(exc),
                source_path=str(yaml_path),
                path=logical_path,
                loc=error_loc_for_yaml_path(logical_path, locations),
            )
        ]
        return _emit_schema_result(yaml_path, schema_path, errors, [], args, ok=False, source_lines=source_lines)

    errors, warnings = _collect_schema_issues(
        yaml_data_dict,
        schema,
        args,
        jsonschema_module,
        source_path=str(yaml_path),
        locations=locations,
    )
    ok = not errors

    return _emit_schema_result(
        yaml_path,
        schema_path,
        errors,
        warnings,
        args,
        ok=ok,
        source_lines=source_lines,
    )


def _load_schema_or_error(
    schema_path: Path,
    *,
    yaml_path: Path,
    args: argparse.Namespace,
) -> Tuple[Optional[Dict[str, Any]], int]:
    if not schema_path.exists():
        _emit_error(
            "Schema 文件不存在: {}".format(schema_path),
            json_output=args.json,
            yaml_path=yaml_path,
            schema_path=schema_path,
            mode="schema-validate",
        )
        return None, 1

    try:
        return _load_json_schema(schema_path), 0
    except json.JSONDecodeError:
        _emit_error(
            "Schema JSON 无法解析: {}".format(schema_path),
            json_output=args.json,
            yaml_path=yaml_path,
            schema_path=schema_path,
            mode="schema-validate",
        )
        return None, 1


def _get_jsonschema_module(
    args: argparse.Namespace,
    *,
    yaml_path: Path,
    schema_path: Path,
) -> Optional[Any]:
    if _HAS_JSONSCHEMA and jsonschema is not None:
        return jsonschema
    _emit_error(
        "缺少 jsonschema 依赖,请安装 scalim[cli]",
        json_output=args.json,
        yaml_path=yaml_path,
        schema_path=schema_path,
        mode="schema-validate",
    )
    return None


def _collect_schema_issues(
    yaml_data: Dict[str, Any],
    schema: Dict[str, Any],
    args: argparse.Namespace,
    jsonschema_module: Any,
    *,
    source_path: str,
    locations: YamlLocationIndex,
) -> Tuple[List[ErrorEnvelope], List[ErrorEnvelope]]:
    try:
        issues = collect_jsonschema_validation_issues(
            yaml_data,
            schema,
            jsonschema_module=jsonschema_module,
            include_context=bool(args.verbose),
            filter_additional_properties=True,
        )
    except ScalimJsonSchemaCollectorError as exc:
        errors = [
            ErrorEnvelope(
                code="yaml_schema_validate_error",
                message=str(exc),
                source_path=source_path,
                path="(root)",
                loc=error_loc_for_yaml_path("(root)", locations),
            )
        ]
        return errors, []

    errors = _issues_to_rows(
        issues,
        source_path=source_path,
        locations=locations,
        default_code="yaml_schema_validate_error",
    )
    errors.extend(
        _find_legacy_field_errors(
            yaml_data,
            source_path=source_path,
            locations=locations,
        )
    )
    errors.extend(
        _issues_to_rows(
            find_unknown_fields(yaml_data, schema),
            source_path=source_path,
            locations=locations,
            default_code="yaml_unknown_field",
        )
    )
    demand_book_ids = _extract_demand_book_ids(yaml_data)
    errors.extend(
        _find_removed_outputs_defaults_errors(
            yaml_data,
            source_path=source_path,
            locations=locations,
            default_code="yaml_schema_validate_error",
        )
    )
    errors.extend(
        _find_demand_book_binding_errors(
            yaml_data,
            source_path=source_path,
            locations=locations,
            available_book_ids=demand_book_ids,
            default_code="yaml_schema_validate_error",
        )
    )
    errors.extend(
        _find_retry_enabled_missing_should_retry_errors(
            yaml_data,
            source_path=source_path,
            locations=locations,
            default_code="yaml_schema_validate_error",
        )
    )
    return errors, []


def _emit_schema_result(
    yaml_path: Path,
    schema_path: Path,
    errors: List[ErrorEnvelope],
    warnings: List[ErrorEnvelope],
    args: argparse.Namespace,
    *,
    ok: bool,
    source_lines: Optional[List[str]],
) -> int:
    if args.json:
        payload = ValidationPayload(
            mode="schema-validate",
            ok=ok,
            yaml_path=str(yaml_path),
            schema_path=str(schema_path),
            errors=errors,
            warnings=warnings,
        )
        _write_line(json.dumps(payload.as_dict(), ensure_ascii=False))
    else:
        _render_result(
            yaml_path,
            errors=errors,
            warnings=warnings,
            verbose=args.verbose,
            source_lines=source_lines,
        )
    return 0 if ok else 1


def _run_schema_show(_args: argparse.Namespace) -> int:
    schema_path = _default_schema_path()
    if not schema_path.exists():
        _emit_error("Schema 文件不存在: {}".format(schema_path), json_output=False)
        return 1
    _write_raw(schema_path.read_text(encoding="utf-8"))
    return 0


def _run_schema_path(_args: argparse.Namespace) -> int:
    schema_path = _default_schema_path()
    if not schema_path.exists():
        _emit_error("Schema 文件不存在: {}".format(schema_path), json_output=False)
        return 1
    _write_line(str(schema_path))
    return 0


def _run_upsert_lsp_comment(args: argparse.Namespace) -> int:
    args_dict = vars(args)
    schema_type = str(args_dict.get("schema_type", yaml_dsl_lsp.DEFAULT_SCHEMA_TYPE) or yaml_dsl_lsp.DEFAULT_SCHEMA_TYPE)
    schema_path = str(args_dict.get("schema_path", yaml_dsl_lsp.DEFAULT_SCHEMA_PATH) or yaml_dsl_lsp.DEFAULT_SCHEMA_PATH)
    comment_style = str(args_dict.get("comment_style", yaml_dsl_lsp.DEFAULT_COMMENT_STYLE) or yaml_dsl_lsp.DEFAULT_COMMENT_STYLE).strip()

    try:
        schema_ref = yaml_dsl_lsp.resolve_schema_ref(schema_type, schema_path)
    except ValueError as exc:
        _emit_error(str(exc), json_output=False)
        return 1

    try:
        schema_modelines = yaml_dsl_lsp.make_schema_modelines(schema_ref, comment_style=comment_style)
    except ValueError as exc:
        _emit_error(str(exc), json_output=False)
        return 1

    exit_code = 0
    changed: List[Path] = []
    unchanged: List[Path] = []

    paths = list(args_dict.get("paths", []) or [])
    for raw_path in paths:
        path = raw_path
        if not path.exists():
            _write_line_stderr("错误: YAML 文件不存在: {}".format(path))
            exit_code = 1
            continue
        if not path.is_file():
            _write_line_stderr("错误: 不是文件: {}".format(path))
            exit_code = 1
            continue

        result = yaml_dsl_lsp.upsert_schema_modelines_file(path, schema_modelines=schema_modelines)
        if result.error:
            _write_line_stderr("错误: {} ({})".format(result.error, path))
            exit_code = 1
            continue
        if result.changed:
            changed.append(path)
            _write_line("UPDATED {}".format(path))
        else:
            unchanged.append(path)
            _write_line("OK {}".format(path))

    if changed or unchanged:
        _write_line("")
        _write_line(
            "Summary: {} updated, {} ok".format(
                len(changed),
                len(unchanged),
            )
        )

    return exit_code


__all__ = []



# region imports

from typing import Callable, Dict, Hashable, List, Optional, Sequence, Set, Tuple

from ..typedefs import FieldValue
from ..vendor.compact.typing_extensionsx import override

# endregion


class _DenseFieldStorage:
    """批次字段的 `Dense` 存储表示: `values` + `present` 掩码.

    说明:
    - `present[i]==1` 表示该行有值(即使值为 `None`),`present[i]==0` 表示缺失.
    - 用 `bytearray` 避免缺失哨兵带来的 `pickle` 兼容性问题.
    """

    values: List[FieldValue]
    present: bytearray
    present_count: int

    def __init__(self, row_count: int) -> None:
        self.values = [None] * int(row_count)
        self.present = bytearray(int(row_count))
        self.present_count = 0


class BatchContext:
    _data: Dict[str, Dict[Hashable, FieldValue]]
    _required_fields: Optional[Set[str]]
    _on_field_set: Optional[Callable[[str, Hashable], None]]
    _disabled_rows: Optional[Set[Hashable]]

    def __init__(
        self,
        required_fields: Optional[Set[str]] = None,
        *,
        on_field_set: Optional[Callable[[str, Hashable], None]] = None,
    ) -> None:
        self._data = {}
        self._required_fields = required_fields
        self._on_field_set = on_field_set
        self._disabled_rows = None

    def set_field_value(self, field_key: str, row_id: Hashable, value: FieldValue) -> None:
        if self._disabled_rows is not None and row_id in self._disabled_rows:
            return

        # 注意: 内存优化 - 剪枝: 只存储需要的字段
        if self._required_fields is not None and field_key not in self._required_fields:
            return

        if field_key not in self._data:
            self._data[field_key] = {}
        self._data[field_key][row_id] = value
        if self._on_field_set is not None:
            self._on_field_set(field_key, row_id)

    def get_field_value(self, field_key: str, row_id: Hashable, default: Optional[FieldValue] = None) -> FieldValue:
        field_data = self._data.get(field_key)
        if field_data is None:
            return default
        return field_data.get(row_id, default)

    def has_field(self, field_key: str) -> bool:
        return field_key in self._data

    def delete_field(self, field_key: str) -> None:
        # 注意: 内存优化 - 及时删除不再需要的字段: 当一个字段的值已经被所有依赖它的字段使用完毕后, 可以调用此方法删除该字段,释放内存
        if field_key in self._data:
            del self._data[field_key]

    def delete_row_from_field(self, field_key: str, row_id: Hashable) -> None:
        """FR023 行级释放: 删除特定行在特定字段中的值.

        `row_id` 为批次内行号(`batch_row_nth`),不是主键.
        触发点:该行在该字段已不再被后续计算/联结使用时调用,用于释放内存.
        """
        field_data = self._data.get(field_key)
        if field_data and row_id in field_data:
            del field_data[row_id]
            if not field_data:
                del self._data[field_key]

    def delete_row_from_all_fields(self, row_id: Hashable, exclude_fields: Optional[Set[str]] = None) -> List[str]:
        """FR023 行级释放: 删除特定行在所有字段中的值,但保留 `exclude_fields` 中的字段.

        `row_id` 为批次内行号(`batch_row_nth`),不是主键.
        触发点:该行已被写出且不再被后续步骤使用时调用,以清理上下文内存.
        """
        exclude = exclude_fields or set()
        released_fields: List[str] = []
        for field_key in list(self._data.keys()):
            if field_key not in exclude:
                field_data = self._data.get(field_key)
                if field_data and row_id in field_data:
                    del field_data[row_id]
                    released_fields.append(field_key)
                    if not field_data:
                        del self._data[field_key]
        return released_fields

    def disable_row(self, row_id: Hashable) -> None:
        if self._disabled_rows is None:
            self._disabled_rows = set()
        self._disabled_rows.add(row_id)

    def get_field_values_for_row(self, row_id: Hashable, field_keys: List[str]) -> Dict[str, FieldValue]:
        # 注意: 内存优化 - 批量获取减少查找开销
        return {key: self.get_field_value(key, row_id) for key in field_keys}

    def clear(self) -> None:
        # 注意: 内存优化 - 批次结束后清空所有数据
        self._data.clear()
        if self._disabled_rows:
            self._disabled_rows.clear()

    def get_all_rows_for_field(self, field_key: str) -> Set[Hashable]:
        field_data = self._data.get(field_key)
        if field_data is None:
            return set()
        return set(field_data.keys())

    def get_field_keys(self) -> Set[str]:
        return set(self._data.keys())

    def get_field_count(self) -> int:
        return len(self._data)


class DenseBatchContext(BatchContext):
    """针对连续整数 `row_id` 批次的 `BatchContext` `Dense` 优化实现.

    仅用于批次内 `row_id` 为连续 `int` 的场景(例如 `pipeline` 生成的 `range` 行号).
    若需要通用 `row_id`,请使用 `BatchContext`.
    """

    _base_row_id: int
    _row_count: int
    _dense_data: Dict[str, _DenseFieldStorage]

    def __init__(
        self,
        *,
        base_row_id: int,
        row_count: int,
        required_fields: Optional[Set[str]] = None,
        on_field_set: Optional[Callable[[str, Hashable], None]] = None,
    ) -> None:
        super(DenseBatchContext, self).__init__(required_fields=required_fields, on_field_set=on_field_set)
        self._base_row_id = int(base_row_id)
        self._row_count = int(max(0, row_count))
        self._dense_data = {}

    def _idx_of(self, row_id: Hashable) -> Optional[int]:
        if not isinstance(row_id, int):
            return None
        idx = int(row_id) - int(self._base_row_id)
        if idx < 0 or idx >= int(self._row_count):
            return None
        return idx

    def _ensure_storage(self, field_key: str) -> _DenseFieldStorage:
        storage = self._dense_data.get(field_key)
        if storage is None:
            storage = _DenseFieldStorage(self._row_count)
            self._dense_data[field_key] = storage
        return storage

    @override
    def set_field_value(self, field_key: str, row_id: Hashable, value: FieldValue) -> None:
        if self._disabled_rows is not None and row_id in self._disabled_rows:
            return
        if self._required_fields is not None and field_key not in self._required_fields:
            return

        idx = self._idx_of(row_id)
        if idx is None:
            # 该实现仅用于连续 `int row_id`;若不满足条件,让调用方回退到通用实现.
            msg = "`DenseBatchContext` row_id 不在范围内或不是 `int`: {!r}".format(row_id)
            raise ValueError(msg)

        storage = self._ensure_storage(field_key)
        if storage.present[idx] == 0:
            storage.present[idx] = 1
            storage.present_count += 1
        storage.values[idx] = value

        if self._on_field_set is not None:
            self._on_field_set(field_key, row_id)

    @override
    def get_field_value(self, field_key: str, row_id: Hashable, default: Optional[FieldValue] = None) -> FieldValue:
        storage = self._dense_data.get(field_key)
        if storage is None:
            return default
        idx = self._idx_of(row_id)
        if idx is None:
            return default
        if storage.present[idx] == 0:
            return default
        return storage.values[idx]

    @override
    def has_field(self, field_key: str) -> bool:
        return field_key in self._dense_data

    @override
    def delete_field(self, field_key: str) -> None:
        if field_key in self._dense_data:
            del self._dense_data[field_key]

    @override
    def delete_row_from_field(self, field_key: str, row_id: Hashable) -> None:
        storage = self._dense_data.get(field_key)
        if storage is None:
            return
        idx = self._idx_of(row_id)
        if idx is None:
            return
        if storage.present[idx] == 0:
            return
        storage.present[idx] = 0
        storage.values[idx] = None
        storage.present_count -= 1
        if storage.present_count <= 0:
            _ = self._dense_data.pop(field_key, None)

    @override
    def delete_row_from_all_fields(self, row_id: Hashable, exclude_fields: Optional[Set[str]] = None) -> List[str]:
        exclude = exclude_fields or set()
        idx = self._idx_of(row_id)
        if idx is None:
            return []

        released_fields: List[str] = []
        for field_key in list(self._dense_data.keys()):
            if field_key in exclude:
                continue
            storage = self._dense_data[field_key]
            if storage.present[idx] == 0:
                continue
            storage.present[idx] = 0
            storage.values[idx] = None
            storage.present_count -= 1
            released_fields.append(field_key)
            if storage.present_count <= 0:
                _ = self._dense_data.pop(field_key, None)
        return released_fields

    @override
    def get_field_values_for_row(self, row_id: Hashable, field_keys: List[str]) -> Dict[str, FieldValue]:
        return {key: self.get_field_value(key, row_id) for key in field_keys}

    @override
    def clear(self) -> None:
        self._dense_data.clear()
        if self._disabled_rows:
            self._disabled_rows.clear()

    @override
    def get_all_rows_for_field(self, field_key: str) -> Set[Hashable]:
        storage = self._dense_data.get(field_key)
        if storage is None:
            return set()
        out: Set[Hashable] = set()
        base = int(self._base_row_id)
        for idx, present in enumerate(storage.present):
            if present:
                out.add(base + idx)
        return out

    @override
    def get_field_keys(self) -> Set[str]:
        return set(self._dense_data.keys())

    @override
    def get_field_count(self) -> int:
        return len(self._dense_data)


def _try_resolve_dense_range(row_ids: Sequence[Hashable]) -> Optional[Tuple[int, int]]:
    if not row_ids:
        return None
    first = row_ids[0]
    if not isinstance(first, int):
        return None
    base = int(first)
    for idx, row_id in enumerate(row_ids):
        if not isinstance(row_id, int):
            return None
        if int(row_id) != base + int(idx):
            return None
    return base, len(row_ids)


def create_batch_context_for_rows(
    row_ids: Sequence[Hashable],
    *,
    required_fields: Optional[Set[str]] = None,
    on_field_set: Optional[Callable[[str, Hashable], None]] = None,
) -> BatchContext:
    resolved = _try_resolve_dense_range(row_ids)
    if resolved is None:
        return BatchContext(required_fields=required_fields, on_field_set=on_field_set)
    base, row_count = resolved
    return DenseBatchContext(base_row_id=base, row_count=row_count, required_fields=required_fields, on_field_set=on_field_set)


__all__ = []



import logging
from collections.abc import Set as AbstractSet
from typing import Any, Callable, Dict, Optional, Set, Tuple

from ..._internal.loggingx import prefix
from ...events import (
    EVENT_ADAPTIVE_SCHEDULER_DECISION,
    EVENT_BATCH_END,
    EVENT_BATCH_START,
    EVENT_COLUMN_WRITE,
    EVENT_DIAGNOSTIC_WARNING,
    EVENT_ERROR,
    EVENT_FIELD_COMPUTE,
    EVENT_FIELD_SLIM,
    EVENT_LOADER_CALL,
    EVENT_LOADER_RETRY,
    EVENT_LOADER_SLIM,
    EVENT_PIPELINE_END,
    EVENT_PIPELINE_START,
    EVENT_RELATION_LOOKUP,
    EVENT_ROW_RELEASE,
    EVENT_ROW_WRITE,
    EVENT_STAGE_SPAN,
    EVENT_WORKFLOW_NODE_CANCELLED,
    EVENT_WORKFLOW_NODE_END,
    EVENT_WORKFLOW_NODE_START,
)

_logger = logging.getLogger(__name__)

HOOK_RAISED_EXCEPTION_WARNING = prefix("hooks") + "钩子 %s.%s 抛出异常"

_HOOK_TYPED_DISPATCH_MAP: Dict[str, str] = {
    EVENT_PIPELINE_START: "on_pipeline_start",
    EVENT_PIPELINE_END: "on_pipeline_end",
    EVENT_BATCH_START: "on_batch_start",
    EVENT_BATCH_END: "on_batch_end",
    EVENT_LOADER_CALL: "on_loader_call",
    EVENT_FIELD_COMPUTE: "on_field_compute",
    EVENT_ERROR: "on_error",
    EVENT_DIAGNOSTIC_WARNING: "on_diagnostic_warning",
    EVENT_FIELD_SLIM: "on_field_slim",
    EVENT_ROW_WRITE: "on_row_write",
    EVENT_ROW_RELEASE: "on_row_release",
    EVENT_LOADER_SLIM: "on_loader_slim",
    EVENT_COLUMN_WRITE: "on_column_write",
}

HOOK_TYPED_DISPATCH_MAP: Dict[str, str] = _HOOK_TYPED_DISPATCH_MAP

CATALOG_EVENT_TYPES: Tuple[str, ...] = (
    EVENT_PIPELINE_START,
    EVENT_PIPELINE_END,
    EVENT_BATCH_START,
    EVENT_BATCH_END,
    EVENT_LOADER_CALL,
    EVENT_LOADER_RETRY,
    EVENT_FIELD_COMPUTE,
    EVENT_ERROR,
    EVENT_DIAGNOSTIC_WARNING,
    EVENT_FIELD_SLIM,
    EVENT_ROW_WRITE,
    EVENT_ROW_RELEASE,
    EVENT_LOADER_SLIM,
    EVENT_COLUMN_WRITE,
    EVENT_RELATION_LOOKUP,
    EVENT_STAGE_SPAN,
    EVENT_ADAPTIVE_SCHEDULER_DECISION,
    EVENT_WORKFLOW_NODE_START,
    EVENT_WORKFLOW_NODE_END,
    EVENT_WORKFLOW_NODE_CANCELLED,
)


def resolve_mro_attr(cls: type, name: str) -> Any:
    for mro_cls in cls.__mro__:
        cls_dict = mro_cls.__dict__
        if name in cls_dict:
            return cls_dict[name]
    return None


def read_optional_attr(obj: Any, name: str) -> Any:
    try:
        return obj.__getattribute__(name)
    except AttributeError:
        return None


def read_callable_attr(obj: Any, name: str) -> Optional[Callable[..., Any]]:
    value = read_optional_attr(obj, name)
    if value is None or not callable(value):
        return None
    return value


def validate_event_types(hook: Any, value: Any) -> Optional[Set[str]]:
    if value is None:
        return None
    if not isinstance(value, AbstractSet):
        msg = "hook.event_types must be None or Set[str]; got {} for {}".format(type(value).__name__, type(hook).__name__)
        raise TypeError(msg)
    validated: Set[str] = set()
    for item in value:
        if not isinstance(item, str):
            msg = "hook.event_types must contain only str; got {} element {!r} for {}".format(
                type(item).__name__, item, type(hook).__name__
            )
            raise TypeError(msg)
        validated.add(item)
    return validated


__all__ = []



from typing import Optional

from ..typedefs import KeyNormalizationMode


def normalize_key_normalization(value: object) -> KeyNormalizationMode:
    if value is None:
        return "raw"
    raw = str(value).strip()
    if raw in ("raw", "auto_str", "force_str"):
        return raw  # type: ignore[return-value]
    msg = "Invalid key_normalization={!r}. Expected 'raw', 'auto_str', or 'force_str'.".format(value)
    raise ValueError(msg)


def should_apply_str_key_normalization(
    mode: KeyNormalizationMode,
    *,
    has_explicit_cast: bool,
) -> bool:
    if mode == "raw":
        return False
    if mode == "auto_str":
        return not bool(has_explicit_cast)
    return True


def is_experimental_key_normalization(mode: Optional[object]) -> bool:
    try:
        resolved = normalize_key_normalization(mode)
    except ValueError:
        return False
    return resolved != "raw"


__all__ = [
    "is_experimental_key_normalization",
    "normalize_key_normalization",
    "should_apply_str_key_normalization",
]



"""`scalim.ob.presets.viz` 稳定导出面.

说明:
- 对外稳定导入路径仍为 `scalim.ob.presets.viz`
- `_internal` 子包属于实现细节(非公共契约)
"""

from .._internal.viz_config import VizObserverConfig
from .._internal.viz_output import VizEventEmitter
from .observer import VizObserver
from .output_composition import augment_viz_graph_snapshot_for_output_composition
from .workflow import WorkflowVizObserver, build_workflow_viz_graph_snapshot

__all__ = [
    "VizEventEmitter",
    "VizObserver",
    "VizObserverConfig",
    "WorkflowVizObserver",
    "augment_viz_graph_snapshot_for_output_composition",
    "build_workflow_viz_graph_snapshot",
]



"""`workflow` 配置类型(稳定导入路径).

说明:
- 该模块提供更稳定、更明确的类型导入路径
- 当前实现仍位于 `workflow_config.py`,此处仅做 `re-export`
- 运行时需兼容 `Python 3.6`
"""

from .workflow_config import (
    ScalimWorkflowConfigError,
    WorkflowCachePoolBudget,
    WorkflowCachePoolOptions,
    WorkflowCachePoolPin,
    WorkflowConfig,
    WorkflowOptions,
    WorkflowRun,
)

__all__ = [
    "ScalimWorkflowConfigError",
    "WorkflowCachePoolBudget",
    "WorkflowCachePoolOptions",
    "WorkflowCachePoolPin",
    "WorkflowConfig",
    "WorkflowOptions",
    "WorkflowRun",
]



"""`DSL` 家族包.

对外推荐入口统一为 `scalim.dsl.<dsl_name>`.

例如:
- `scalim.dsl.by_yaml`: `YAML` `DSL` 官方入口
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import by_yaml  # noqa: TC004

__all__ = ("by_yaml",)



# region imports

import json
import logging
import random
import time
from pathlib import Path
from typing import Any, Dict, List, Optional

from ..._internal.loggingx import get_logger, prefix
from ...events._events import RelationLookupEvent
from ...typedefs import RelationLookupResult, RelationReportFormat
from ...vendor.compact.typing_extensionsx import override
from ...vendor.dataclassesx import asdict, dataclass, field
from .._internal.console_report import emit_info, emit_warning, format_percent
from ..observer import EventDispatchObserver

# endregion

_LOGGER = get_logger("relations")


@dataclass
class RelationSample:
    row_id: Any
    fk_raw: Any
    fk_normalized: Any
    target_source: str
    result: RelationLookupResult
    timestamp: float = field(default_factory=time.time)
    fk_type: Optional[str] = None
    expected_type: Optional[str] = None
    error_message: Optional[str] = None


@dataclass
class RelationSourceStats:
    total_lookups: int = 0
    hit_count: int = 0
    miss_count: int = 0
    null_key_count: int = 0
    type_mismatch_count: int = 0

    @property
    def hit_rate(self) -> float:
        denominator = self.hit_count + self.miss_count
        if denominator == 0:
            return 0.0
        return self.hit_count / denominator


@dataclass
class RelationMetrics:
    """关联命中统计口径."""

    total_lookups: int = 0
    hit_count: int = 0
    miss_count: int = 0
    null_key_count: int = 0
    type_mismatch_count: int = 0
    per_source_stats: Dict[str, RelationSourceStats] = field(default_factory=dict)
    samples: List[RelationSample] = field(default_factory=list)
    type_mismatch_samples: List[RelationSample] = field(default_factory=list)

    @property
    def hit_rate(self) -> float:
        denominator = self.hit_count + self.miss_count
        if denominator == 0:
            return 0.0
        return self.hit_count / denominator

    def get_source_stats(self, source_id: str) -> RelationSourceStats:
        if source_id not in self.per_source_stats:
            self.per_source_stats[source_id] = RelationSourceStats()
        return self.per_source_stats[source_id]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "summary": {
                "total_lookups": self.total_lookups,
                "hit_count": self.hit_count,
                "miss_count": self.miss_count,
                "hit_rate": round(self.hit_rate, 4),
                "null_key_count": self.null_key_count,
                "type_mismatch_count": self.type_mismatch_count,
            },
            "sources": {source_id: asdict(stats) for source_id, stats in self.per_source_stats.items()},
            "samples": [asdict(s) for s in self.samples],
            "type_mismatch_samples": [asdict(s) for s in self.type_mismatch_samples],
        }

    def to_json(self, indent: int = 2) -> str:
        # 注意: `samples` 可能包含非 `JSON` 原生类型(例如 `Decimal`/`datetime`/`tuple key` 等来自用户数据).
        # 这里用 `default=str` 保证可观测性输出足够稳健.
        return json.dumps(self.to_dict(), ensure_ascii=False, indent=indent, default=str)


@dataclass
class RelationConfig:
    enabled: bool = True
    sampling_rate: float = 0.01
    log_type_mismatch: bool = True
    max_samples: int = 1000
    report_format: RelationReportFormat = "console"
    output_path: Optional[str] = None
    include_details: bool = False
    logger: logging.Logger = field(default=_LOGGER)

    @classmethod
    def default(cls) -> "RelationConfig":
        return cls()


class RelationObserver(EventDispatchObserver):
    config: RelationConfig
    metrics: RelationMetrics

    def __init__(
        self,
        config: Optional[RelationConfig] = None,
    ) -> None:
        if config is None:
            config = RelationConfig.default()
        self.config = config
        self.metrics = RelationMetrics()

    def on_relation_lookup(self, event: RelationLookupEvent) -> None:  # noqa: C901
        if not self.config.enabled:
            return

        m = self.metrics
        m.total_lookups += 1
        source_stats = m.get_source_stats(event.target_source)
        source_stats.total_lookups += 1

        if event.result == "hit":
            m.hit_count += 1
            source_stats.hit_count += 1
        elif event.result == "miss":
            m.miss_count += 1
            source_stats.miss_count += 1
        elif event.result == "null_key":
            m.null_key_count += 1
            source_stats.null_key_count += 1
        elif event.result == "type_error":
            m.type_mismatch_count += 1
            source_stats.type_mismatch_count += 1

        if self.config.sampling_rate < 1.0 and random.random() > self.config.sampling_rate:  # noqa: S311
            return

        sample = RelationSample(
            row_id=event.row_id,
            fk_raw=event.fk_raw,
            fk_normalized=event.fk_normalized,
            target_source=event.target_source,
            result=event.result,
            fk_type=event.fk_type,
            expected_type=event.expected_type,
            error_message=event.error_message,
        )

        if event.result == "type_error":
            if self.config.log_type_mismatch:
                emit_warning(
                    self.config.logger,
                    "relations",
                    "type_error",
                    row_id=event.row_id,
                    fk_raw=event.fk_raw,
                    target_source=event.target_source,
                    error=event.error_message,
                )
            if len(m.type_mismatch_samples) < self.config.max_samples:
                m.type_mismatch_samples.append(sample)
        elif len(m.samples) < self.config.max_samples:
            m.samples.append(sample)

    def record_lookup(
        self,
        row_id: Any,
        fk_raw: Any,
        fk_normalized: Any,
        target_source: str,
        result: RelationLookupResult,
        fk_type: Optional[str] = None,
        expected_type: Optional[str] = None,
        error_message: Optional[str] = None,
    ) -> None:
        event = RelationLookupEvent(
            field_key="",
            row_id=row_id,
            fk_raw=fk_raw,
            fk_normalized=fk_normalized,
            target_source=target_source,
            result=result,
            fk_type=fk_type,
            expected_type=expected_type,
            error_message=error_message,
        )
        self.on_relation_lookup(event)

    def get_metrics(self) -> RelationMetrics:
        return self.metrics

    def on_pipeline_start(self, event: Any) -> None:
        _ = event
        self.reset()

    def on_pipeline_end(self, event: Any) -> None:
        _ = event
        self._output_report()

    def print_summary(self) -> None:
        m = self.metrics
        logger = self.config.logger

        emit_info(
            logger,
            "relations",
            "summary",
            total_lookups=int(m.total_lookups),
            hits=int(m.hit_count),
            misses=int(m.miss_count),
            null_keys=int(m.null_key_count),
            type_errors=int(m.type_mismatch_count),
            hit_rate=format_percent(m.hit_rate, digits=2),
        )

        if m.per_source_stats:
            for source_id in sorted(m.per_source_stats.keys()):
                stats = m.per_source_stats[source_id]
                emit_info(
                    logger,
                    "relations",
                    "per_source",
                    source=str(source_id),
                    total=int(stats.total_lookups),
                    hits=int(stats.hit_count),
                    misses=int(stats.miss_count),
                    null_keys=int(stats.null_key_count),
                    type_errors=int(stats.type_mismatch_count),
                    hit_rate=format_percent(stats.hit_rate, digits=2),
                )

        if m.type_mismatch_samples:
            showing = min(5, len(m.type_mismatch_samples))
            emit_info(logger, "relations", "type_mismatch_samples", total=len(m.type_mismatch_samples), showing=int(showing))
            for sample in m.type_mismatch_samples[:showing]:
                emit_info(
                    logger,
                    "relations",
                    "type_mismatch_sample",
                    row_id=sample.row_id,
                    fk_raw=sample.fk_raw,
                    fk_type=sample.fk_type,
                    expected_type=sample.expected_type,
                    target_source=sample.target_source,
                    error=sample.error_message,
                )

    def _write_json_report(self) -> None:
        output_path = self.config.output_path
        if not output_path:
            self.config.logger.info("\n%s", json.dumps(self._build_report_dict(), ensure_ascii=False, indent=2, default=str))
            return

        try:
            path = Path(output_path)
            if path.parent and not path.parent.exists():
                path.parent.mkdir(parents=True, exist_ok=True)
            _ = path.write_text(json.dumps(self._build_report_dict(), ensure_ascii=False, indent=2, default=str), encoding="utf-8")
            self.config.logger.info("%s报告已写入: %s", prefix("relations"), output_path)
        except OSError as e:
            self.config.logger.warning("%s写入报告失败: %s", prefix("relations"), e)

    def _build_report_dict(self) -> Dict[str, Any]:
        return self.metrics.to_dict()

    def _output_report(self) -> None:
        fmt = self.config.report_format
        if fmt == "none":
            return
        if fmt == "console":
            self.print_summary()
        elif fmt == "json":
            self._write_json_report()

    @override
    def close(self) -> None:
        self._output_report()

    def reset(self) -> None:
        self.metrics = RelationMetrics()


__all__ = [
    "RelationConfig",
    "RelationMetrics",
    "RelationObserver",
    "RelationSample",
    "RelationSourceStats",
]



"""执行契约(`DSL` 无关).

说明:
- 本模块仅包含纯 `contracts`(数据结构),不包含执行编排逻辑.
- 运行时需兼容 `Python 3.6`.
"""

from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Union

from ..typedefs import KeyNormalizationMode, ParallelMode, RowData
from ..vendor.dataclassesx import dataclass
from ..vendor.dataclassesx import field as dataclass_field
from .output_contracts import ExportLayout, OutputSpec

if TYPE_CHECKING:
    from ..hooks import IExecutionHook
    from ..ob.observer import Observer
    from ..ob.presets.viz import VizObserverConfig
    from ..planning.plan import ExecutionPlan
    from ..sinks import InMemoryCsv, ISink
    from ..sinks.rows import InMemoryRows
    from ..spec.ir import DemandIr
    from .guardrails import GuardrailsPolicy
    from .loader_retry import LoaderRetryPolicies
    from .output_composition import OutputCompositionSpec, OutputTargetStats


@dataclass(frozen=True)
class ObservabilitySpec:
    """与 `DSL` 无关的可观测性请求(用于运行编排).

    - `viz_config` 可选;在构建执行计划后会物化为 `VizObserver`.
    """

    fallback_logger_enabled: bool = False
    viz_config: Optional["VizObserverConfig"] = None


@dataclass(frozen=True)
class ExecutionRequest:
    export_layout: ExportLayout
    """导出布局(字段顺序与可选表头)."""

    output: OutputSpec = dataclass_field(default_factory=OutputSpec)
    """输出策略(例如输出格式、路径、编码、是否流式)."""

    sink: Optional["ISink"] = None
    """可选:显式指定输出端;若为 `None` 则按 `output` 策略创建."""

    observability: Optional[ObservabilitySpec] = None
    """可选:可观测性请求(例如 `viz` 配置)."""

    components: Optional[List[Union["Observer", "IExecutionHook"]]] = None
    """可选:要挂载的 `Observer`/`Hook` 组件列表."""

    batch_size: Optional[int] = None
    """可选:覆盖批大小(`None` 表示不覆盖)."""

    parallel_mode: ParallelMode = "seq"
    """并行模式(`seq` 或 `adaptive`)."""

    max_workers: int = 0
    """最大并发工作数提示(`0` 表示自动)."""

    key_normalization: KeyNormalizationMode = "raw"
    """可选: `key` 规范化模式(实验性;默认 `raw`)."""

    guardrails: Optional["GuardrailsPolicy"] = None
    """可选:运行时护栏策略."""

    loader_retry: Optional["LoaderRetryPolicies"] = None
    """可选:加载重试策略."""

    output_composition: Optional["OutputCompositionSpec"] = None
    """可选:多输出组合请求(`IR/Python-only`).

    当提供该字段时:
    - `output`/`sink` 的单输出装配将被忽略
    - 运行计划的目标字段将由组合请求的 `required_demand_fields` 计算得出
    """

    main_rows: Optional[Iterable[RowData]] = None
    """可选:显式注入 `main_rows`(当提供时绕过主数据源 `loader`)."""

    capture_in_memory_rows: bool = False
    """可选:捕获本次运行输出的 `InMemoryRows`(保留 `FieldValue` 类型域;默认关闭)."""


@dataclass(frozen=True)
class ExecutionResult:
    """与 `DSL` 无关的执行结果.

    注意:
    - `total_rows` 统计写入到实际输出端的行数(包括 `NullSink`),这是输出/写出的行数.
    - 可观测性指标可能使用不同口径来做低开销吞吐估算
      (例如 `PerformanceMetrics.total_rows` 统计的是输入 `row_ids`).
    """

    output_path: Optional[str]
    total_rows: int
    duration: float
    demand_ir: "DemandIr"
    plan: "ExecutionPlan"
    outputs: Optional[Dict[str, str]] = None
    """可选:输出目标到 `output_path` 的映射(多输出组合时提供)."""

    output_target_stats: Optional[List["OutputTargetStats"]] = None
    """可选:每个输出目标的统计(行数/耗时/错误/禁用)(多输出组合时提供)."""

    in_memory_csv_outputs: Optional[Dict[str, "InMemoryCsv"]] = None
    """可选: `workflow-managed` 无路径 CSV 输出的内存中间态(多输出组合时提供)."""

    in_memory_rows: Optional["InMemoryRows"] = None
    """可选: `workflow-intermediate-store` 的 `InMemoryRows` 中间态(显式启用时提供)."""


__all__ = [
    "ExecutionRequest",
    "ExecutionResult",
    "ObservabilitySpec",
]



"""`YAML DSL` 运行时编译器.

本模块将 `YAML DSL` 文件转换为:
- 已校验的配置(`DemandConfig`)
- 中间表示(`DemandIr`)
- 运行时请求(`ExecutionRequest`)
"""

import os
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, FrozenSet, List, Mapping, Optional, Sequence, Set, Tuple, cast

from ....execution.guardrails import (
    GuardrailMode,
    GuardrailsComputePolicy,
    GuardrailsLoaderPolicy,
    GuardrailsPolicy,
    GuardrailsRelationsPolicy,
)
from ....execution.loader_retry import LoaderRetryPolicies, LoaderRetryPoliciesSpec, LoaderRetryPolicy, LoaderRetryPolicySpec
from ....execution.run_ir import ExecutionRequest, ObservabilitySpec, OutputSpec, export_layout_from_demand_ir
from ....spec.ir import DemandIr
from ....vendor.dataclassesx import replace
from .._internal.config_parsing.loader import YamlDemandLoader
from .._internal.config_parsing.template_precompile import DEFAULT_RENDERED_YAML_MAX_LEN
from ..diagnostics import format_duplicate_effective_field_display_names_message
from ..init_var_nodes import parse_init_var_mapping_node
from ..reference_syntax import BUILTIN_CALLABLE_REFERENCE_PREFIX
from ..schema_dsl.constants import (
    DEFAULT_OUTPUT_ENCODING,
    DEFAULT_OUTPUT_HEADER_BY,
    DEFAULT_OUTPUT_INCLUDE_HEADER,
)
from ..schema_dsl.models import (
    BookBudgetConfig,
    BookConfig,
    BookExportXlsxConfig,
    BookWriteDefaultsConfig,
    DemandConfig,
    FileConfig,
    GuardrailsConfig,
    LoaderRetryConfig,
    OutputTargetConfig,
    OutputToConfig,
    OutputWriteConfig,
    ResourcesConfig,
)
from ..schema_dsl.output_enums import (
    BOOK_KINDS,
    BOOK_WRITE_ALIGN_BY_ENUM,
    BOOK_WRITE_HEADER_POLICY_ENUM,
    BOOK_WRITE_MODE_ENUM,
    BOOK_WRITE_ON_CONFLICT_ENUM,
    BOOK_WRITE_ON_MISMATCH_ENUM,
    DEFAULT_BOOK_WRITE_ALIGN_BY,
    DEFAULT_BOOK_WRITE_HEADER_POLICY,
    DEFAULT_BOOK_WRITE_MODE,
    DEFAULT_BOOK_WRITE_ON_CONFLICT,
    DEFAULT_BOOK_WRITE_ON_MISMATCH,
    FILE_KINDS,
)
from .builtin_callables import parse_builtin_callable_id
from .contracts import (
    BookBudgetOverride,
    BookExportXlsxOverride,
    BookResourceOverride,
    BookWriteDefaultsOverride,
    Compilation,
    FileResourceOverride,
    OutputOverride,
    OutputsDefaultsOverride,
    OutputToOverride,
    OutputWriteOverride,
    ResolverTrustedMode,
    ResourcesOverride,
    RunOptions,
    UnsetType,
)
from .conversion import ConfigToIRConverter
from .errors import ALLOWLIST_REQUIRED_MSG, ScalimAllowlistRequiredError, ScalimResolverError
from .observability import compile_observability_spec
from .output_composition_yaml import compile_output_composition_from_yaml
from .references import SecurePythonReferenceResolver, derive_base_module_path

if TYPE_CHECKING:
    from ....execution.output_composition import OutputCompositionSpec


def _ensure_allowlist(
    allowed_modules: FrozenSet[str],
    allowed_functions: Optional[FrozenSet[str]],
) -> None:
    if not allowed_modules and not allowed_functions:
        # 安全审计:仅允许从 `allowlist` 中指定的模块/函数加载,避免 `YAML` 触发任意导入执行.
        raise ScalimAllowlistRequiredError(ALLOWLIST_REQUIRED_MSG)


def validate_allowlist(
    *,
    allowed_modules: FrozenSet[str],
    allowed_functions: Optional[FrozenSet[str]],
) -> None:
    _ensure_allowlist(allowed_modules, allowed_functions)


_OUTPUT_NAME_PATTERN = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_]*$")
_OUTPUT_HEADER_BY_ENUM: Tuple[str, ...] = ("field_id", "name")


def _parse_overrides_outputs_defaults_book_id(defaults: Optional[object], *, path: str) -> Optional[str]:
    if defaults is None:
        return None
    if not isinstance(defaults, OutputsDefaultsOverride):
        msg = "{} must be an OutputsDefaultsOverride".format(path)
        raise TypeError(msg)
    book_id = str(defaults.to.book or "").strip()
    if not book_id:
        msg = "{}.to.book is required".format(path)
        raise ValueError(msg)
    return book_id


def _apply_default_book_binding_to_outputs(
    outputs: Tuple[OutputTargetConfig, ...],
    *,
    default_book_id: str,
) -> Tuple[OutputTargetConfig, ...]:
    if not outputs:
        return outputs
    if not default_book_id:
        return outputs

    updated: List[OutputTargetConfig] = []
    for out_cfg in outputs:
        to_cfg = out_cfg.to
        if to_cfg is None:
            updated.append(replace(out_cfg, to=OutputToConfig(book=str(default_book_id))))
            continue

        file_id = str(to_cfg.file or "").strip() if to_cfg.file is not None else ""
        book_id = str(to_cfg.book or "").strip() if to_cfg.book is not None else ""
        if file_id or book_id:
            updated.append(out_cfg)
            continue

        updated.append(replace(out_cfg, to=replace(to_cfg, book=str(default_book_id))))

    return tuple(updated)


def _parse_typed_overrides_output_to(raw: OutputToOverride) -> OutputToConfig:
    file_id = str(raw.file or "").strip() if raw.file is not None else None
    book_id = str(raw.book or "").strip() if raw.book is not None else None
    sheet = str(raw.sheet or "").strip() if raw.sheet is not None else None
    if file_id == "":
        file_id = None
    if book_id == "":
        book_id = None
    if sheet == "":
        sheet = None

    return OutputToConfig(file=file_id, book=book_id, sheet=sheet)


def _parse_typed_overrides_output_write(raw: OutputWriteOverride, *, path: str) -> OutputWriteConfig:
    def _as_opt_str(value: Optional[str]) -> Optional[str]:
        if value is None:
            return None
        v = str(value).strip()
        return v or None

    include_header = raw.include_header
    if include_header is not None and not isinstance(include_header, bool):
        msg = "{}.include_header must be a boolean".format(path)
        raise TypeError(msg)

    header_fields_output_by = _as_opt_str(raw.header_fields_output_by)
    if header_fields_output_by is not None and header_fields_output_by not in _OUTPUT_HEADER_BY_ENUM:
        msg = "{}.header_fields_output_by={!r} is invalid; expected one of: {}".format(
            path, header_fields_output_by, ", ".join(_OUTPUT_HEADER_BY_ENUM)
        )
        raise ValueError(msg)

    return OutputWriteConfig(
        include_header=include_header,
        mode=_as_opt_str(raw.mode),
        align_by=_as_opt_str(raw.align_by),
        header_policy=_as_opt_str(raw.header_policy),
        header_fields_output_by=header_fields_output_by,
        on_mismatch=_as_opt_str(raw.on_mismatch),
        on_conflict=_as_opt_str(raw.on_conflict),
    )


def _parse_overrides_outputs_targets(  # noqa: C901, PLR0912, PLR0915
    overrides: Sequence[OutputOverride],
    demand_ir: DemandIr,
    *,
    path: str,
    default_book_id: Optional[str],
    default_book_ref: str,
) -> Tuple[OutputTargetConfig, ...]:
    if not isinstance(overrides, tuple) and not isinstance(overrides, list):
        msg = "{} must be a sequence of OutputOverride".format(path)
        raise TypeError(msg)
    if not overrides:
        msg = "{} cannot be empty".format(path)
        raise ValueError(msg)

    known_field_ids: Set[str] = {str(fid) for fid in demand_ir.fields}
    seen_names: Set[str] = set()
    parsed: List[OutputTargetConfig] = []

    for idx, item in enumerate(overrides):
        if not isinstance(item, OutputOverride):
            msg = "{}.{} must be an OutputOverride".format(path, idx)
            raise TypeError(msg)

        name = str(item.name or "").strip()
        if not name:
            msg = "{}.{}.name is required".format(path, idx)
            raise ValueError(msg)
        if not _OUTPUT_NAME_PATTERN.match(name):
            msg = "{}.{}.name={!r} is invalid; expected identifier like [a-zA-Z_][a-zA-Z0-9_]*".format(path, idx, name)
            raise ValueError(msg)
        if name in seen_names:
            msg = "{} has duplicate output name: {}".format(path, name)
            raise ValueError(msg)
        seen_names.add(name)

        fields = item.fields
        if not isinstance(fields, tuple):
            msg = "{}.{}.fields must be a tuple[str, ...]".format(path, idx)
            raise TypeError(msg)
        if not fields:
            msg = "{}.{}.fields must not be empty".format(path, idx)
            raise ValueError(msg)

        field_ids: List[str] = []
        for field_idx, field_id_raw in enumerate(fields):
            if not isinstance(field_id_raw, str):
                msg = "{}.{}.fields.{} must be a field_id string".format(path, idx, field_idx)
                raise TypeError(msg)
            field_id = field_id_raw.strip()
            if not field_id:
                msg = "{}.{}.fields.{} must not be empty".format(path, idx, field_idx)
                raise ValueError(msg)
            field_ids.append(field_id)

        unknown_fields = [fid for fid in field_ids if fid not in known_field_ids]
        if unknown_fields:
            msg = "{}.{}.fields reference unknown fields: {}".format(path, idx, ", ".join(sorted(set(unknown_fields))))
            raise ValueError(msg)

        to_override = item.to
        if not isinstance(to_override, OutputToOverride):
            msg = "{}.{}.to must be an OutputToOverride".format(path, idx)
            raise TypeError(msg)
        to_cfg = _parse_typed_overrides_output_to(to_override)

        write_cfg = None
        if item.write is not None:
            write_override = item.write
            if not isinstance(write_override, OutputWriteOverride):
                msg = "{}.{}.write must be an OutputWriteOverride".format(path, idx)
                raise TypeError(msg)
            write_cfg = _parse_typed_overrides_output_write(write_override, path="{}.{}.write".format(path, idx))

        file_id = str(to_cfg.file or "").strip() if to_cfg.file is not None else ""
        book_id = str(to_cfg.book or "").strip() if to_cfg.book is not None else ""
        sheet = str(to_cfg.sheet or "").strip() if to_cfg.sheet is not None else ""

        if file_id:
            if book_id:
                msg = "{}.{}.to must declare exactly one of to.file or to.book".format(path, idx)
                raise ValueError(msg)
            if sheet:
                msg = "{}.{}.to.sheet is not allowed with to.file".format(path, idx)
                raise ValueError(msg)
        else:
            effective_book_id = book_id or str(default_book_id or "").strip()
            if not effective_book_id:
                msg = ("Missing output destination for {}.{}.to; set {}.{}.to.book explicitly or provide {}.").format(
                    path, idx, path, idx, default_book_ref
                )
                raise ValueError(msg)
            if book_id != effective_book_id:
                to_cfg = replace(to_cfg, book=str(effective_book_id))
                book_id = effective_book_id

        if file_id and write_cfg is not None:
            invalid_keys: List[str] = []
            if write_cfg.mode is not None:
                invalid_keys.append("mode")
            if write_cfg.align_by is not None:
                invalid_keys.append("align_by")
            if write_cfg.header_policy is not None:
                invalid_keys.append("header_policy")
            if write_cfg.on_mismatch is not None:
                invalid_keys.append("on_mismatch")
            if write_cfg.on_conflict is not None:
                invalid_keys.append("on_conflict")
            if invalid_keys:
                msg = "{}.{}.write.{} only apply to book outputs".format(path, idx, ", ".join(invalid_keys))
                raise ValueError(msg)
        if book_id and write_cfg is not None and str(write_cfg.mode or "").strip() == "append" and write_cfg.include_header is not None:
            msg = "{}.{}.write.include_header is not allowed for append-mode book outputs; use write.header_policy".format(path, idx)
            raise ValueError(msg)

        parsed.append(
            OutputTargetConfig(
                name=str(name),
                from_=None,
                to=to_cfg,
                write=write_cfg,
                fields=tuple(field_ids),
                where=None,
                aggregate=None,
                requires=(),
            )
        )

    return tuple(parsed)


def _output_requires_unique_effective_field_display_names(  # noqa: C901
    config: DemandConfig, output: OutputTargetConfig
) -> bool:
    to_cfg = output.to
    if to_cfg is None:
        return False

    write_cfg = output.write
    header_by = str(DEFAULT_OUTPUT_HEADER_BY)
    if write_cfg is not None and write_cfg.header_fields_output_by is not None:
        header_by = str(write_cfg.header_fields_output_by)
    if header_by != "name":
        return False

    file_id = str(to_cfg.file or "").strip() if to_cfg.file is not None else ""
    if file_id:
        include_header = DEFAULT_OUTPUT_INCLUDE_HEADER
        if write_cfg is not None and write_cfg.include_header is not None:
            include_header = bool(write_cfg.include_header)
        return bool(include_header)

    book_id = str(to_cfg.book or "").strip() if to_cfg.book is not None else ""
    if not book_id:
        return False

    book = None
    if config.resources is not None:
        book = config.resources.books.get(book_id)

    mode = DEFAULT_BOOK_WRITE_MODE
    header_policy = DEFAULT_BOOK_WRITE_HEADER_POLICY
    if book is not None and book.write_defaults is not None:
        mode = str(book.write_defaults.mode or DEFAULT_BOOK_WRITE_MODE)
        header_policy = str(book.write_defaults.header_policy or DEFAULT_BOOK_WRITE_HEADER_POLICY)
    if write_cfg is not None and write_cfg.mode is not None:
        mode = str(write_cfg.mode)
    if write_cfg is not None and write_cfg.header_policy is not None:
        header_policy = str(write_cfg.header_policy)

    if mode == "append":
        return header_policy != "never"

    include_header = DEFAULT_OUTPUT_INCLUDE_HEADER
    if write_cfg is not None and write_cfg.include_header is not None:
        include_header = bool(write_cfg.include_header)
    return bool(include_header)


def _should_validate_unique_effective_field_display_names(config: DemandConfig, outputs: Tuple[OutputTargetConfig, ...]) -> bool:
    if not bool(config.validate_unique_field_names):
        return False
    return any(_output_requires_unique_effective_field_display_names(config, t) for t in outputs)


def _validate_unique_effective_field_display_names(demand_ir: DemandIr) -> None:
    conflicts: Dict[str, List[str]] = {}
    for field_id, field_ir in demand_ir.fields.items():
        name = str(field_ir.name or "").strip()
        effective = name or str(field_id)
        conflicts.setdefault(effective, []).append(str(field_id))

    duplicates = {name: ids for name, ids in conflicts.items() if len(ids) > 1}
    if not duplicates:
        return

    msg = format_duplicate_effective_field_display_names_message(duplicates)
    raise ValueError(msg)


def _as_guardrail_mode(value: str) -> GuardrailMode:
    if value == "quiet":
        return "quiet"
    if value == "fast_fail":
        return "fast_fail"
    msg = "Invalid guardrail mode: '{}'".format(value)
    raise ValueError(msg)


def _compile_guardrails_policy(config: Optional[GuardrailsConfig]) -> GuardrailsPolicy:
    if config is None:
        return GuardrailsPolicy.disabled()

    loader = config.loader
    relations = config.relations
    compute = config.compute

    return GuardrailsPolicy(
        enabled=bool(config.enabled),
        mode=_as_guardrail_mode(str(config.mode)),
        loader=GuardrailsLoaderPolicy(
            validate_result=bool(loader.validate_result) if loader is not None else False,
            required_fields=tuple(str(item) for item in (loader.required_fields or ())) if loader is not None else (),
            on_transform_error=_as_guardrail_mode(str(loader.on_transform_error))
            if (loader is not None and loader.on_transform_error)
            else None,
        ),
        relations=GuardrailsRelationsPolicy(
            null_key_max_rate=relations.null_key_max_rate if relations is not None else None,
            type_error_max_rate=relations.type_error_max_rate if relations is not None else None,
        ),
        compute=GuardrailsComputePolicy(
            on_error=_as_guardrail_mode(str(compute.on_error)) if (compute is not None and compute.on_error) else None,
        ),
    )


def _retry_spec_from_yaml(
    config: Optional[LoaderRetryConfig],
    *,
    resolver: SecurePythonReferenceResolver,
) -> LoaderRetryPolicySpec:
    if config is None:
        return LoaderRetryPolicySpec()
    should_retry = None
    if config.should_retry:
        should_retry = cast("Any", resolver.resolve(str(config.should_retry)))  # pragma: allow-cast resolver callable signature boundary
    return LoaderRetryPolicySpec(
        enabled=bool(config.enabled) if config.enabled is not None else None,
        should_retry=should_retry,
        max_attempts=int(config.max_attempts) if config.max_attempts is not None else None,
        max_elapsed_seconds=float(config.max_elapsed_seconds) if config.max_elapsed_seconds is not None else None,
        backoff=str(config.backoff) if config.backoff is not None else None,
        base_delay_seconds=float(config.base_delay_seconds) if config.base_delay_seconds is not None else None,
        max_delay_seconds=float(config.max_delay_seconds) if config.max_delay_seconds is not None else None,
        jitter=bool(config.jitter) if config.jitter is not None else None,
    )


def _merge_retry_specs(base: LoaderRetryPolicySpec, override: Optional[LoaderRetryPolicySpec]) -> LoaderRetryPolicySpec:
    if override is None:
        return base
    return LoaderRetryPolicySpec(
        enabled=override.enabled if override.enabled is not None else base.enabled,
        should_retry=override.should_retry if override.should_retry is not None else base.should_retry,
        max_attempts=override.max_attempts if override.max_attempts is not None else base.max_attempts,
        max_elapsed_seconds=override.max_elapsed_seconds if override.max_elapsed_seconds is not None else base.max_elapsed_seconds,
        backoff=override.backoff if override.backoff is not None else base.backoff,
        base_delay_seconds=override.base_delay_seconds if override.base_delay_seconds is not None else base.base_delay_seconds,
        max_delay_seconds=override.max_delay_seconds if override.max_delay_seconds is not None else base.max_delay_seconds,
        jitter=override.jitter if override.jitter is not None else base.jitter,
    )


def _finalize_retry_policy(spec: LoaderRetryPolicySpec, *, base: Optional[LoaderRetryPolicy] = None) -> LoaderRetryPolicy:
    base_policy = base or LoaderRetryPolicy.disabled()
    enabled = spec.enabled if spec.enabled is not None else base_policy.enabled
    should_retry = spec.should_retry if spec.should_retry is not None else base_policy.should_retry
    max_attempts = spec.max_attempts if spec.max_attempts is not None else base_policy.max_attempts
    max_elapsed = spec.max_elapsed_seconds if spec.max_elapsed_seconds is not None else base_policy.max_elapsed_seconds
    backoff = spec.backoff if spec.backoff is not None else base_policy.backoff
    base_delay = spec.base_delay_seconds if spec.base_delay_seconds is not None else base_policy.base_delay_seconds
    max_delay = spec.max_delay_seconds if spec.max_delay_seconds is not None else base_policy.max_delay_seconds
    jitter = spec.jitter if spec.jitter is not None else base_policy.jitter

    if enabled and should_retry is None:
        msg = "loader_retry.enabled=true requires should_retry (provide in YAML or via driver injection)"
        raise ValueError(msg)

    return LoaderRetryPolicy(
        enabled=bool(enabled),
        should_retry=should_retry,
        max_attempts=int(max_attempts),
        max_elapsed_seconds=float(max_elapsed),
        backoff=str(backoff),
        base_delay_seconds=float(base_delay),
        max_delay_seconds=float(max_delay),
        jitter=bool(jitter),
    )


def _compile_loader_retry_policies(
    config: DemandConfig,
    *,
    resolver: SecurePythonReferenceResolver,
    overrides: Optional[LoaderRetryPoliciesSpec],
) -> Optional[LoaderRetryPolicies]:
    yaml_global = _retry_spec_from_yaml(config.retry, resolver=resolver)
    driver_global = overrides.default if overrides is not None else None
    global_spec = _merge_retry_specs(yaml_global, driver_global)

    base_policy = LoaderRetryPolicy.disabled()
    global_policy = _finalize_retry_policy(global_spec, base=base_policy)

    known_loaders = set(config.sources.keys())
    known_loaders.add(config.main_source.source_id)
    if overrides is not None and overrides.by_loader:
        unknown = set(overrides.by_loader.keys()) - known_loaders
        if unknown:
            msg = "Unknown loader_retry.by_loader keys: {}".format(", ".join(sorted(unknown)))
            raise ValueError(msg)

    by_loader: Dict[str, LoaderRetryPolicy] = {}

    main_source_id = config.main_source.source_id
    main_yaml = _retry_spec_from_yaml(config.main_source.retry, resolver=resolver)
    main_driver = overrides.by_loader.get(main_source_id) if overrides is not None else None
    main_spec = _merge_retry_specs(_merge_retry_specs(global_spec, main_yaml), main_driver)
    main_policy = _finalize_retry_policy(main_spec, base=base_policy)
    if main_policy != global_policy:
        by_loader[main_source_id] = main_policy

    for source_id, source_config in config.sources.items():
        src_yaml = _retry_spec_from_yaml(source_config.retry, resolver=resolver)
        src_driver = overrides.by_loader.get(source_id) if overrides is not None else None
        src_spec = _merge_retry_specs(_merge_retry_specs(global_spec, src_yaml), src_driver)
        src_policy = _finalize_retry_policy(src_spec, base=base_policy)
        if src_policy != global_policy:
            by_loader[source_id] = src_policy

    if not global_policy.enabled and not by_loader:
        return None

    return LoaderRetryPolicies(default=global_policy, by_loader=by_loader)


def load_config(
    yaml_path: str,
    *,
    template_vars: Optional[Mapping[str, object]] = None,
    template_sandbox: str = "safe",
    rendered_yaml_max_len: int = DEFAULT_RENDERED_YAML_MAX_LEN,
    allowed_yaml_roots: Optional[Sequence[str]] = None,
) -> DemandConfig:
    loader = YamlDemandLoader()
    return loader.load(
        yaml_path,
        template_vars=template_vars,
        template_sandbox=template_sandbox,
        rendered_yaml_max_len=rendered_yaml_max_len,
        allowed_yaml_roots=allowed_yaml_roots,
    )


def _normalize_builtin_callable_id(id_raw: object, *, label: str) -> str:
    builtin_id = str(id_raw or "").strip()
    if not builtin_id:
        msg = "{}: <id> must not be empty".format(label)
        raise ValueError(msg)
    if builtin_id.startswith(BUILTIN_CALLABLE_REFERENCE_PREFIX):
        msg = "{}: <id> must not include prefix '{}': {!r}".format(label, BUILTIN_CALLABLE_REFERENCE_PREFIX, builtin_id)
        raise ValueError(msg)
    try:
        _ = parse_builtin_callable_id(BUILTIN_CALLABLE_REFERENCE_PREFIX + builtin_id)
    except ScalimResolverError as exc:
        msg = "{}: invalid <id> {!r}: {}".format(label, builtin_id, exc)
        raise ValueError(msg) from exc
    return builtin_id


def _compile_builtin_callable_vocab_value(
    builtin_id: str,
    value_raw: object,
    *,
    trusted_resolver: SecurePythonReferenceResolver,
) -> Callable[..., Any]:
    if callable(value_raw):
        return cast("Callable[..., Any]", value_raw)  # pragma: allow-cast callable() typed narrowing

    if isinstance(value_raw, str):
        reference = value_raw.strip()
        if not reference:
            msg = "builtin_callables[{!r}]: value must not be empty".format(builtin_id)
            raise ValueError(msg)
        if reference.startswith(BUILTIN_CALLABLE_REFERENCE_PREFIX):
            msg = "builtin_callables[{!r}]: value must be a Python reference, not builtin '^<id>': {!r}".format(builtin_id, value_raw)
            raise ValueError(msg)
        try:
            return trusted_resolver.resolve(reference)
        except ScalimResolverError as exc:
            msg = "builtin_callables[{!r}]: failed to resolve Python reference {!r}: {}".format(builtin_id, reference, exc)
            raise ValueError(msg) from exc

    msg = "builtin_callables[{!r}]: expected a callable or Python reference string, got: {}".format(builtin_id, type(value_raw).__name__)
    raise TypeError(msg)


def _compile_builtin_callables_vocab(
    builtin_callables: Optional[Mapping[str, object]],
    *,
    allowed_modules: Optional[FrozenSet[str]] = None,
    allowed_functions: Optional[FrozenSet[str]] = None,
) -> Optional[Dict[str, Callable[..., Any]]]:
    if builtin_callables is None:
        return None

    compiled: Dict[str, Callable[..., Any]] = {}
    trusted_resolver = SecurePythonReferenceResolver(
        allowed_modules=allowed_modules,
        allowed_functions=allowed_functions,
        base_module_path=None,
    )

    for builtin_id_raw, value_raw in builtin_callables.items():
        builtin_id = _normalize_builtin_callable_id(builtin_id_raw, label="builtin_callables")
        compiled[builtin_id] = _compile_builtin_callable_vocab_value(
            builtin_id,
            value_raw,
            trusted_resolver=trusted_resolver,
        )

    return compiled


def _validate_public_builtin_callable_ids(public_ids: Optional[Sequence[str]]) -> Optional[Tuple[str, ...]]:
    if public_ids is None:
        return None

    validated: List[str] = []
    for raw in public_ids:
        validated.append(_normalize_builtin_callable_id(raw, label="public_builtin_callable_ids"))
    return tuple(validated)


def create_reference_resolver(
    *,
    allowed_modules: FrozenSet[str],
    allowed_functions: Optional[FrozenSet[str]],
    resolver_trusted_mode: ResolverTrustedMode = ResolverTrustedMode.STRICT_ALLOWLIST,
    base_module_path: Optional[str] = None,
    builtin_callables: Optional[Mapping[str, object]] = None,
    public_builtin_callable_ids: Optional[Sequence[str]] = None,
) -> SecurePythonReferenceResolver:
    compiled_builtin_callables = _compile_builtin_callables_vocab(
        builtin_callables,
        allowed_modules=allowed_modules,
        allowed_functions=allowed_functions,
    )
    validated_public_ids = _validate_public_builtin_callable_ids(public_builtin_callable_ids)
    return SecurePythonReferenceResolver(
        allowed_modules=allowed_modules,
        allowed_functions=allowed_functions,
        resolver_trusted_mode=resolver_trusted_mode,
        base_module_path=base_module_path,
        builtin_callables_by_id=compiled_builtin_callables,
        public_builtin_callable_ids=validated_public_ids,
    )


def compile_ir(
    config: DemandConfig,
    *,
    resolver: SecurePythonReferenceResolver,
    init_vars: Optional[Dict[str, object]] = None,
) -> DemandIr:
    converter = ConfigToIRConverter(resolver=resolver, init_vars=init_vars)
    return converter.convert(config)


def _resolve_effective_outputs_and_path(
    config: DemandConfig,
    demand_ir: DemandIr,
    *,
    options: RunOptions,
) -> Tuple[Tuple[OutputTargetConfig, ...], str]:
    overrides = options.overrides
    overrides_outputs = overrides.outputs if overrides is not None else None
    default_book_id = _parse_overrides_outputs_defaults_book_id(
        None if overrides is None else overrides.outputs_defaults,
        path="overrides.outputs_defaults",
    )
    if overrides_outputs is not None:
        return (
            _parse_overrides_outputs_targets(
                overrides_outputs,
                demand_ir,
                path="overrides.outputs",
                default_book_id=default_book_id,
                default_book_ref="overrides.outputs_defaults.to.book",
            ),
            "overrides.outputs",
        )
    if config.outputs:
        outputs = tuple(config.outputs)
        if default_book_id is not None:
            outputs = _apply_default_book_binding_to_outputs(outputs, default_book_id=str(default_book_id))
        return outputs, "outputs"
    return (), "outputs"


def _normalize_non_empty_pathlike_value(raw: object, *, path: str) -> str:
    if raw is None:
        msg = "{} is required".format(path)
        raise ValueError(msg)
    if isinstance(raw, os.PathLike):
        value = str(os.fspath(raw)).strip()
    elif isinstance(raw, str):
        value = str(raw).strip()
    else:
        msg = "{} must be a string or os.PathLike".format(path)
        raise TypeError(msg)
    if not value:
        msg = "{} is required".format(path)
        raise ValueError(msg)
    return value


def _parse_non_empty_path_or_init_var(raw: object, *, path: str) -> Any:
    if isinstance(raw, dict):
        return {"$init_var": parse_init_var_mapping_node(cast("Dict[str, Any]", raw), path=path)}  # pragma: allow-cast dict narrowing
    if raw is None:
        msg = "{} is required".format(path)
        raise ValueError(msg)
    if not isinstance(raw, os.PathLike) and not isinstance(raw, str):
        msg = "{} must be a non-empty string or {{$init_var: <name>}}".format(path)
        raise TypeError(msg)
    return _normalize_non_empty_pathlike_value(raw, path=path)


def _parse_optional_path_or_init_var(raw: object, *, path: str) -> Any:
    if raw is None:
        return None
    if isinstance(raw, dict):
        return {"$init_var": parse_init_var_mapping_node(cast("Dict[str, Any]", raw), path=path)}  # pragma: allow-cast dict narrowing
    if isinstance(raw, os.PathLike):
        value = str(os.fspath(raw)).strip()
        if not value:
            msg = "{} must not be empty".format(path)
            raise ValueError(msg)
        return value
    if isinstance(raw, str):
        value = raw.strip()
        if not value:
            msg = "{} must not be empty".format(path)
            raise ValueError(msg)
        return value
    msg = "{} must be a string or {{$init_var: <name>}}".format(path)
    raise TypeError(msg)


def _normalize_override_mapping_key(raw_id: object, *, path: str) -> str:
    if not isinstance(raw_id, str) or not str(raw_id).strip():
        msg = "{} keys must be non-empty strings".format(path)
        raise ValueError(msg)
    return str(raw_id).strip()


def _overlay_book_write_defaults_override(
    base: Optional[BookWriteDefaultsConfig],
    override: BookWriteDefaultsOverride,
    *,
    path: str,
) -> BookWriteDefaultsConfig:
    base_defaults = base or BookWriteDefaultsConfig(
        mode=str(DEFAULT_BOOK_WRITE_MODE),
        align_by=str(DEFAULT_BOOK_WRITE_ALIGN_BY),
        header_policy=str(DEFAULT_BOOK_WRITE_HEADER_POLICY),
        on_mismatch=str(DEFAULT_BOOK_WRITE_ON_MISMATCH),
        on_conflict=str(DEFAULT_BOOK_WRITE_ON_CONFLICT),
    )

    mode = str(base_defaults.mode or DEFAULT_BOOK_WRITE_MODE) if override.mode is None else str(override.mode).strip()
    align_by = str(base_defaults.align_by or DEFAULT_BOOK_WRITE_ALIGN_BY) if override.align_by is None else str(override.align_by).strip()
    header_policy = (
        str(base_defaults.header_policy or DEFAULT_BOOK_WRITE_HEADER_POLICY)
        if override.header_policy is None
        else str(override.header_policy).strip()
    )
    on_mismatch = (
        str(base_defaults.on_mismatch or DEFAULT_BOOK_WRITE_ON_MISMATCH)
        if override.on_mismatch is None
        else str(override.on_mismatch).strip()
    )
    on_conflict = (
        str(base_defaults.on_conflict or DEFAULT_BOOK_WRITE_ON_CONFLICT)
        if override.on_conflict is None
        else str(override.on_conflict).strip()
    )

    validations = (
        ("mode", mode, BOOK_WRITE_MODE_ENUM),
        ("align_by", align_by, BOOK_WRITE_ALIGN_BY_ENUM),
        ("header_policy", header_policy, BOOK_WRITE_HEADER_POLICY_ENUM),
        ("on_mismatch", on_mismatch, BOOK_WRITE_ON_MISMATCH_ENUM),
        ("on_conflict", on_conflict, BOOK_WRITE_ON_CONFLICT_ENUM),
    )
    for key, value, allowed in validations:
        if value not in allowed:
            msg = "Invalid write_defaults.{}={!r}; expected one of: {}".format(str(key), value, ", ".join(allowed))
            err = "{} (path={}.{})".format(msg, path, str(key))
            raise ValueError(err)

    return BookWriteDefaultsConfig(
        mode=str(mode),
        align_by=str(align_by),
        header_policy=str(header_policy),
        on_mismatch=str(on_mismatch),
        on_conflict=str(on_conflict),
    )


def _overlay_book_budget_override(
    base: Optional[BookBudgetConfig],
    override: BookBudgetOverride,
    *,
    path: str,
) -> BookBudgetConfig:
    if base is None:
        if override.max_sheets is None or override.max_total_cells is None:
            msg = "{} requires max_sheets and max_total_cells when creating a new xlsx_memory book".format(path)
            raise ValueError(msg)
        max_sheets = override.max_sheets
        max_total_cells = override.max_total_cells
    else:
        max_sheets = base.max_sheets if override.max_sheets is None else override.max_sheets
        max_total_cells = base.max_total_cells if override.max_total_cells is None else override.max_total_cells

    if isinstance(max_sheets, bool) or not isinstance(max_sheets, int):
        msg = "{}.max_sheets must be an integer".format(path)
        raise TypeError(msg)
    if isinstance(max_total_cells, bool) or not isinstance(max_total_cells, int):
        msg = "{}.max_total_cells must be an integer".format(path)
        raise TypeError(msg)
    if int(max_sheets) < 1:
        msg = "{}.max_sheets must be >= 1".format(path)
        raise ValueError(msg)
    if int(max_total_cells) < 1:
        msg = "{}.max_total_cells must be >= 1".format(path)
        raise ValueError(msg)

    return BookBudgetConfig(max_sheets=int(max_sheets), max_total_cells=int(max_total_cells))


def _overlay_book_export_xlsx_override(
    base: Optional[BookExportXlsxConfig],
    override: BookExportXlsxOverride,
    *,
    path: str,
) -> BookExportXlsxConfig:
    if base is None:
        if override.path is None:
            msg = "{}.path is required when creating export_xlsx".format(path)
            raise ValueError(msg)
        export_path = _parse_non_empty_path_or_init_var(override.path, path="{}.path".format(path))
        return BookExportXlsxConfig(
            path=export_path,
            write_lock=bool(override.write_lock) if override.write_lock is not None else False,
            allow_formulas=bool(override.allow_formulas) if override.allow_formulas is not None else False,
        )

    export_path_any: Any = base.path
    if override.path is not None:
        export_path_any = _parse_non_empty_path_or_init_var(override.path, path="{}.path".format(path))

    write_lock = base.write_lock if override.write_lock is None else bool(override.write_lock)
    allow_formulas = base.allow_formulas if override.allow_formulas is None else bool(override.allow_formulas)
    return BookExportXlsxConfig(
        path=export_path_any,
        write_lock=bool(write_lock),
        allow_formulas=bool(allow_formulas),
    )


def _apply_book_override(  # noqa: C901, PLR0912, PLR0915
    base: Optional[BookConfig],
    override: BookResourceOverride,
    *,
    path: str,
) -> BookConfig:
    kind = str(base.kind or "").strip() if base is not None else ""
    book_path: Any = base.path if base is not None else None
    budget = base.budget if base is not None else None
    export_xlsx = base.export_xlsx if base is not None else None
    allow_formulas = bool(base.allow_formulas) if base is not None else False
    write_lock = bool(base.write_lock) if base is not None else False
    write_defaults = base.write_defaults if base is not None else None

    if override.kind is not None:
        kind = str(override.kind or "").strip()
        if not kind:
            msg = "{}.kind must be a non-empty string".format(path)
            raise ValueError(msg)
    if kind not in BOOK_KINDS:
        msg = "{}.kind={!r} is invalid; expected one of: {}".format(path, kind, ", ".join(BOOK_KINDS))
        raise ValueError(msg)

    if override.path is not None:
        book_path = _parse_optional_path_or_init_var(override.path, path="{}.path".format(path))
    if override.allow_formulas is not None:
        if not isinstance(override.allow_formulas, bool):
            msg = "{}.allow_formulas must be a bool".format(path)
            raise TypeError(msg)
        allow_formulas = bool(override.allow_formulas)
    if override.write_lock is not None:
        if not isinstance(override.write_lock, bool):
            msg = "{}.write_lock must be a bool".format(path)
            raise TypeError(msg)
        write_lock = bool(override.write_lock)

    if override.budget is not None:
        budget = _overlay_book_budget_override(budget, override.budget, path="{}.budget".format(path))
    if override.export_xlsx is not None:
        export_xlsx = _overlay_book_export_xlsx_override(export_xlsx, override.export_xlsx, path="{}.export_xlsx".format(path))
    if override.write_defaults is not None:
        write_defaults = _overlay_book_write_defaults_override(
            write_defaults, override.write_defaults, path="{}.write_defaults".format(path)
        )

    if kind == "xlsx_file":
        if override.budget is not None or override.export_xlsx is not None:
            msg = "{}.budget/export_xlsx are not allowed for kind=xlsx_file".format(path)
            raise ValueError(msg)
        if book_path is None:
            msg = "{}.path is required for kind=xlsx_file".format(path)
            raise ValueError(msg)
        if budget is not None:
            msg = "{}.budget is not allowed for kind=xlsx_file".format(path)
            raise ValueError(msg)
        if export_xlsx is not None:
            msg = "{}.export_xlsx is not allowed for kind=xlsx_file".format(path)
            raise ValueError(msg)

    if kind == "xlsx_memory":
        if override.path is not None or override.allow_formulas is not None or override.write_lock is not None:
            msg = "{}.path/allow_formulas/write_lock are not allowed for kind=xlsx_memory (use export_xlsx.*)".format(path)
            raise ValueError(msg)
        if budget is None:
            msg = "{}.budget is required for kind=xlsx_memory".format(path)
            raise ValueError(msg)
        if book_path is not None:
            msg = "{}.path is not allowed for kind=xlsx_memory".format(path)
            raise ValueError(msg)
        if allow_formulas:
            msg = "{}.allow_formulas is not allowed for kind=xlsx_memory".format(path)
            raise ValueError(msg)
        if write_lock:
            msg = "{}.write_lock is not allowed for kind=xlsx_memory".format(path)
            raise ValueError(msg)

    return BookConfig(
        kind=str(kind),
        path=book_path,
        budget=budget,
        export_xlsx=export_xlsx,
        allow_formulas=bool(allow_formulas),
        write_lock=bool(write_lock),
        write_defaults=write_defaults,
    )


def _apply_file_override(base: Optional[FileConfig], override: FileResourceOverride, *, path: str) -> FileConfig:
    kind = str(base.kind or "").strip() if base is not None else ""
    file_path: Any = base.path if base is not None else None
    encoding = str(base.encoding or DEFAULT_OUTPUT_ENCODING) if base is not None else DEFAULT_OUTPUT_ENCODING

    if override.kind is not None:
        kind = str(override.kind or "").strip()
        if not kind:
            msg = "{}.kind must be a non-empty string".format(path)
            raise ValueError(msg)
    if kind not in FILE_KINDS:
        msg = "{}.kind={!r} is invalid; expected one of: {}".format(path, kind, ", ".join(FILE_KINDS))
        raise ValueError(msg)

    if override.path is not None:
        file_path = _parse_optional_path_or_init_var(override.path, path="{}.path".format(path))
    if file_path is None:
        msg = "{}.path is required for kind=csv_file".format(path)
        raise ValueError(msg)

    if override.encoding is not None:
        if not isinstance(override.encoding, str):
            msg = "{}.encoding must be a string".format(path)
            raise TypeError(msg)
        encoding = str(override.encoding).strip() or DEFAULT_OUTPUT_ENCODING

    return FileConfig(kind=str(kind), path=file_path, encoding=str(encoding))


def _apply_resources_override(config: DemandConfig, override: ResourcesOverride) -> DemandConfig:
    if not override.books and not override.files:
        return config

    base_resources = config.resources
    merged_books: Dict[str, BookConfig] = dict(base_resources.books) if base_resources is not None else {}
    merged_files: Dict[str, FileConfig] = dict(base_resources.files) if base_resources is not None else {}

    if override.books:
        for raw_book_id, book_override in override.books.items():
            book_id = _normalize_override_mapping_key(raw_book_id, path="overrides.resources.books")
            if not isinstance(book_override, BookResourceOverride):
                msg = "overrides.resources.books.{} must be a BookResourceOverride".format(book_id)
                raise TypeError(msg)
            merged_books[book_id] = _apply_book_override(
                merged_books.get(book_id),
                book_override,
                path="overrides.resources.books.{}".format(book_id),
            )

    if override.files:
        for raw_file_id, file_override in override.files.items():
            file_id = _normalize_override_mapping_key(raw_file_id, path="overrides.resources.files")
            if not isinstance(file_override, FileResourceOverride):
                msg = "overrides.resources.files.{} must be a FileResourceOverride".format(file_id)
                raise TypeError(msg)
            merged_files[file_id] = _apply_file_override(
                merged_files.get(file_id),
                file_override,
                path="overrides.resources.files.{}".format(file_id),
            )

    return replace(config, resources=ResourcesConfig(books=merged_books, files=merged_files))


def _apply_io_overrides(config: DemandConfig, *, options: RunOptions) -> DemandConfig:
    overrides = options.overrides
    if overrides is None:
        return config

    next_config = config

    resources_override = overrides.resources
    if resources_override is not None:
        next_config = _apply_resources_override(next_config, resources_override)

    return next_config


def _compile_output_composition_for_outputs(
    config: DemandConfig,
    demand_ir: DemandIr,
    *,
    effective_outputs: Tuple[OutputTargetConfig, ...],
    outputs_path: str,
    yaml_base_dir: str,
    options: RunOptions,
    resolver: SecurePythonReferenceResolver,
) -> Optional["OutputCompositionSpec"]:
    if not effective_outputs:
        return None

    if _should_validate_unique_effective_field_display_names(config, effective_outputs):
        _validate_unique_effective_field_display_names(demand_ir)

    config_for_outputs = config if effective_outputs == tuple(config.outputs) else replace(config, outputs=effective_outputs)
    return cast(  # pragma: allow-cast output composition typed narrowing
        "OutputCompositionSpec",
        compile_output_composition_from_yaml(
            config_for_outputs,
            demand_ir,
            resolver=resolver,
            init_vars=options.init_vars,
            yaml_base_dir=str(yaml_base_dir),
            workflow_managed_output_ids=options.workflow_managed_output_ids,
            outputs_path=outputs_path,
            skip_extra_sheets_without_workbook=options.overrides is not None and options.overrides.outputs is not None,
        ),
    )


def build_request(
    config: DemandConfig,
    demand_ir: DemandIr,
    *,
    yaml_base_dir: str,
    options: RunOptions,
    resolver: SecurePythonReferenceResolver,
) -> ExecutionRequest:
    effective_config = _apply_io_overrides(config, options=options)
    effective_outputs, outputs_path = _resolve_effective_outputs_and_path(effective_config, demand_ir, options=options)
    output_composition = _compile_output_composition_for_outputs(
        effective_config,
        demand_ir,
        effective_outputs=effective_outputs,
        outputs_path=outputs_path,
        yaml_base_dir=str(yaml_base_dir),
        options=options,
        resolver=resolver,
    )

    # 无 `outputs` 时走默认策略: 不写文件,但 `sink` 仍可捕获数据.
    export_layout = export_layout_from_demand_ir(demand_ir, (), header_fields_output_by="field_id")
    output_spec = OutputSpec(path=None)
    if output_composition is None:
        target_field_ids = list(demand_ir.fields.keys())
        export_layout = export_layout_from_demand_ir(
            demand_ir,
            target_field_ids,
            header_fields_output_by="field_id",
        )

    observability: Optional[ObservabilitySpec] = None
    components = list(options.components or [])
    if effective_config.observability is not None:
        observability, observers = compile_observability_spec(effective_config.observability)
        components.extend(observers)

    if options.overrides is not None:
        viz_config = options.overrides.viz_config
        if not isinstance(viz_config, UnsetType):
            viz_override = viz_config
            if viz_override is None:
                if observability is not None:
                    observability = replace(observability, viz_config=None)
            elif observability is None:
                observability = ObservabilitySpec(viz_config=viz_override)
            else:
                observability = replace(observability, viz_config=viz_override)
    if not components:
        components = None

    guardrails = options.guardrails or _compile_guardrails_policy(effective_config.guardrails)
    loader_retry = _compile_loader_retry_policies(effective_config, resolver=resolver, overrides=options.loader_retry)
    batch_size = options.batch_size if options.batch_size is not None else effective_config.batch_size

    return ExecutionRequest(
        export_layout=export_layout,
        output=output_spec,
        sink=options.sink,
        output_composition=output_composition,
        observability=observability,
        guardrails=guardrails,
        loader_retry=loader_retry,
        components=components,
        batch_size=batch_size,
        parallel_mode=options.parallel_mode,
        max_workers=options.max_workers,
        key_normalization=options.key_normalization,
    )


def compile(  # noqa: A001
    yaml_path: str,
    *,
    options: RunOptions,
) -> Compilation:
    validate_allowlist(allowed_modules=options.allowed_modules, allowed_functions=options.allowed_functions)
    config = load_config(
        yaml_path,
        template_vars=options.template_vars,
        template_sandbox=options.template_sandbox,
        rendered_yaml_max_len=options.rendered_yaml_max_len,
        allowed_yaml_roots=options.allowed_yaml_roots,
    )
    base_module_path = None
    if _config_uses_relative_references(config):
        base_module_path = derive_base_module_path(yaml_path)
    resolver = create_reference_resolver(
        allowed_modules=options.allowed_modules,
        allowed_functions=options.allowed_functions,
        resolver_trusted_mode=options.resolver_trusted_mode,
        base_module_path=base_module_path,
        builtin_callables=options.builtin_callables,
        public_builtin_callable_ids=options.public_builtin_callable_ids,
    )
    demand_ir = compile_ir(config, resolver=resolver, init_vars=options.init_vars)
    yaml_base_dir = str(Path(str(yaml_path)).expanduser().resolve(strict=False).parent)
    request = build_request(config, demand_ir, yaml_base_dir=yaml_base_dir, options=options, resolver=resolver)
    return Compilation(
        config=config,
        demand_ir=demand_ir,
        request=request,
    )


__all__ = [
    "build_request",
    "compile",
    "compile_ir",
    "create_reference_resolver",
    "load_config",
    "validate_allowlist",
]


def _is_relative_reference(value: Optional[str]) -> bool:
    raw = str(value or "").strip()
    return bool(raw) and raw.startswith(".")


def _config_uses_relative_references(config: DemandConfig) -> bool:
    if _is_relative_reference(config.main_source.loader):
        return True

    if any(
        _is_relative_reference(source.loader) or (source.retry is not None and _is_relative_reference(source.retry.should_retry))
        for source in config.sources.values()
    ):
        return True

    if config.retry is not None and _is_relative_reference(config.retry.should_retry):
        return True
    if config.main_source.retry is not None and _is_relative_reference(config.main_source.retry.should_retry):
        return True

    return any(derived.call_by is not None and _is_relative_reference(derived.call_by) for derived in config.derived_fields.values())



# 本文件由 `scripts/gen-project-constants.py` 自动生成,请勿手改.
# 数据来源: `pyproject.toml` 的 `([project] + [tool.scalim.*])`.

CLI_NAME = "scalim-cli"
DIST_NAME = "scalim"
ENV_BENCH_MAX_WORKERS = "SCALIM_BENCH_MAX_WORKERS"
ENV_BENCH_SCALE = "SCALIM_BENCH_SCALE"
ENV_BENCH_SCOPE = "SCALIM_BENCH_SCOPE"
ENV_SYNC_ANALYSIS_SCRIPT = "SCALIM_ANALYSIS_SCRIPT"
ENV_SYNC_ANALYZE_ARGS = "SCALIM_ANALYZE_ARGS"
ENV_SYNC_ANALYZE_ARGS_PREFIX = "SCALIM_ANALYZE_ARGS_"
ENV_SYNC_DEST_DIR = "SCALIM_DEST_DIR"
ENV_SYNC_LOG_DEST_DIR = "SCALIM_LOG_DEST_DIR"
ENV_SYNC_LOG_REMOTE_DIR = "SCALIM_LOG_REMOTE_DIR"
ENV_SYNC_REMOTE_HOST = "SCALIM_REMOTE_HOST"
ENV_SYNC_REPORT_ROOT_DIR = "SCALIM_REPORT_ROOT_DIR"
ENV_SYNC_TMP_REMOTE_DIR = "SCALIM_TMP_REMOTE_DIR"
IMPORT_ROOT = "scalim"
VERSION = "0.6.0"
VIZ_ARTIFACTS_ROOT = "artifacts/scalim-viz"
VIZ_DIR_NAME = "scalim-viz"
VIZ_REPLAY_ROUTE = "/__scalim_viz__/replay"

__all__ = []



import hashlib
import json
import threading
from collections import OrderedDict
from typing import Callable, Dict, FrozenSet, List, Mapping, Optional, Set, Tuple, overload

from .._internal.utils.json_like import ensure_json_like as _ensure_json_like_ssot
from ..events import (
    EVENT_DIAGNOSTIC_WARNING,
    EVENT_WORKFLOW_CACHE_ACQUIRE,
    EVENT_WORKFLOW_CACHE_EVICT,
    EVENT_WORKFLOW_CACHE_RELEASE,
)
from ..events._events import (
    DiagnosticWarningEvent,
    WorkflowCacheAcquireEvent,
    WorkflowCacheEvictEvent,
    WorkflowCacheReleaseEvent,
)
from ..exceptions import ScalimExecutionError
from ..ob.hub import InstrumentationHub
from ..spec.ir import SourceIr
from ..spec.ir._workflow import WorkflowCachePoolIr
from ..typedefs import LoaderCallKwargs, LoaderResultMapping
from ..vendor.compact.typing_extensionsx import TypeGuard
from ..vendor.dataclassesx import dataclass, field


class ScalimWorkflowCachePoolError(ScalimExecutionError):
    path: str

    def __init__(self, message: str, *, path: str) -> None:
        super(ScalimWorkflowCachePoolError, self).__init__(str(message))
        self.path = str(path or "")


def _is_list(value: object) -> TypeGuard[List[object]]:
    return isinstance(value, list)


def _is_dict(value: object) -> TypeGuard[Dict[object, object]]:
    return isinstance(value, dict)


def _ensure_json_like(value: object, *, path: str) -> object:
    return _ensure_json_like_ssot(
        value,
        path=path,
        value_name="Signature value",
        allowed_types_desc="None/bool/int/float/str/list/tuple/dict[str, ...]",
        dict_key_desc="str",
        require_nonempty_dict_key=False,
        error_cls=ScalimWorkflowCachePoolError,
    )


@overload
def _normalize_json_like(value: Dict[str, object]) -> Dict[str, object]: ...


@overload
def _normalize_json_like(value: List[object]) -> List[object]: ...


@overload
def _normalize_json_like(value: object) -> object: ...


def _normalize_json_like(value: object) -> object:
    if value is None or isinstance(value, (bool, int, float, str)):
        return value
    if _is_list(value):
        return [_normalize_json_like(item) for item in value]
    if _is_dict(value):
        out: Dict[str, object] = {}
        for raw_key in sorted(value.keys(), key=str):
            out[str(raw_key)] = _normalize_json_like(value[raw_key])
        return out
    return value


def _canonical_json_dumps(value: object) -> str:
    return json.dumps(
        value,
        ensure_ascii=False,
        sort_keys=True,
        separators=(",", ":"),
    )


def _format_callable_reference(fn: object) -> str:
    module = str(getattr(fn, "__module__", "") or "").strip()  # pragma: allow-dynattr metadata: callable metadata
    qualname = str(getattr(fn, "__qualname__", "") or "").strip()  # pragma: allow-dynattr metadata: callable metadata
    name = str(getattr(fn, "__name__", "") or "").strip()  # pragma: allow-dynattr metadata: callable metadata

    if not module or module == "builtins":
        return qualname or name or repr(fn)
    if qualname and "." in qualname and "<locals>" not in qualname:
        return "{}:{}".format(module, qualname)
    entry = qualname or name
    if entry:
        return "{}.{}".format(module, entry)
    return module


def _lookup_cast_signature(cast_fn: object) -> Optional[Dict[str, object]]:
    if cast_fn is None:
        return None
    name = getattr(cast_fn, "scalim_lookup_cast_name", None)  # pragma: allow-dynattr plugin: lookup cast metadata
    if isinstance(name, str) and name.strip():
        payload: Dict[str, object] = {"name": str(name)}
        meta = getattr(cast_fn, "scalim_lookup_cast_meta", None)  # pragma: allow-dynattr plugin: lookup cast metadata
        if _is_dict(meta):
            for k, v in meta.items():
                if k == "name":
                    continue
                if isinstance(k, str) and k.strip():
                    payload[str(k)] = _ensure_json_like(v, path="(lookup_cast)")
        return payload

    # 兜底: 尽量保持稳定且 `JSON-safe`.
    return {
        "callable": _format_callable_reference(cast_fn),
    }


@dataclass(frozen=True)
class WorkflowCacheEntrySignature:
    kind: str
    source_id: str
    loader_ref: str
    rendered_params: object
    normalize: Optional[Dict[str, object]] = None
    key: object = None
    lookup_cast: Optional[Dict[str, object]] = None

    def logical_key(self) -> Tuple[str, str]:
        return (str(self.kind), str(self.source_id))

    def as_dict(self) -> Dict[str, object]:
        return {
            "kind": str(self.kind),
            "source_id": str(self.source_id),
            "loader_ref": str(self.loader_ref),
            "rendered_params": self.rendered_params,
            "normalize": self.normalize or None,
            "key": self.key,
            "lookup_cast": self.lookup_cast or None,
        }

    def canonical_key(self) -> str:
        normalized = _normalize_json_like(self.as_dict())
        return _canonical_json_dumps(normalized)

    def digest(self) -> str:
        payload = self.canonical_key().encode("utf-8")
        return hashlib.sha256(payload).hexdigest()[:16]


def build_preload_forever_signature(source: SourceIr, *, rendered_params: LoaderCallKwargs) -> WorkflowCacheEntrySignature:
    params_path = "sources.{}.params".format(source.source_id)
    params = _normalize_json_like(_ensure_json_like(rendered_params, path=params_path))

    normalize_dict: Optional[Dict[str, object]] = None
    if source.normalize is not None:
        norm = source.normalize
        normalize_payload: Dict[str, object] = {
            "kind": norm.kind,
            "key_field": norm.key_field,
            "on_conflict": norm.on_conflict,
            "on_empty": norm.on_empty,
            "on_missing": norm.on_missing,
            "fields": [
                {
                    "name": rule.name,
                    "from_key": rule.from_key,
                    "extract_expr": rule.extract_expr,
                    "extract_segments": list(rule.extract_segments or ()),
                }
                for rule in norm.fields
            ],
        }
        _ = _ensure_json_like(normalize_payload, path="sources.{}.normalize".format(source.source_id))
        normalize_dict = _normalize_json_like(normalize_payload)

    signature = WorkflowCacheEntrySignature(
        kind="preload_forever",
        source_id=str(source.source_id),
        loader_ref=_format_callable_reference(source.loader_spec.callable),
        rendered_params=params,
        normalize=normalize_dict,
        key=_normalize_json_like(_ensure_json_like(source.key.key, path="sources.{}.key".format(source.source_id))),
        lookup_cast=_lookup_cast_signature(source.key.cast),
    )
    # 校验顶层 `JSON-like` 结构.
    _ = _ensure_json_like(signature.as_dict(), path="(signature)")
    return signature


def diff_signature_fields(left: WorkflowCacheEntrySignature, right: WorkflowCacheEntrySignature) -> List[str]:
    diff: List[str] = []
    left_dict = left.as_dict()
    right_dict = right.as_dict()
    for key in sorted(left_dict.keys()):
        if left_dict.get(key) != right_dict.get(key):
            diff.append(str(key))
    return diff


@dataclass
class _CacheEntry:
    signature: WorkflowCacheEntrySignature
    value: Optional[LoaderResultMapping] = None
    loading: bool = False
    lock: threading.Lock = field(default_factory=threading.Lock)


class WorkflowCachePool:
    _workflow_exec_id: str
    _instrumentation: InstrumentationHub
    _conflict_policy: str
    _release_policy: str
    _max_entries: int
    _over_budget_policy: str
    _pinned_logical_keys: FrozenSet[Tuple[str, str]]
    _logical_keys_by_node_id: Mapping[str, FrozenSet[Tuple[str, str]]]
    _remaining_consumers_by_logical_key: Dict[Tuple[str, str], Set[str]]
    _entries: "OrderedDict[str, _CacheEntry]"
    _signature_keys_by_logical_key: Dict[Tuple[str, str], Set[str]]
    _acquired_by_node_id: Dict[str, Set[str]]
    _done_node_ids: Set[str]
    _lock: threading.Lock

    def __init__(
        self,
        *,
        workflow_exec_id: str,
        instrumentation: InstrumentationHub,
        config: WorkflowCachePoolIr,
        logical_keys_by_node_id: Mapping[str, FrozenSet[Tuple[str, str]]],
        consumers_by_logical_key: Mapping[Tuple[str, str], FrozenSet[str]],
    ) -> None:
        self._workflow_exec_id = str(workflow_exec_id)
        self._instrumentation = instrumentation
        self._conflict_policy = str(config.conflict_policy)
        self._release_policy = str(config.release_policy)
        self._max_entries = int(config.budget.max_entries)
        self._over_budget_policy = str(config.budget.over_budget_policy)
        self._pinned_logical_keys = frozenset((str(p.kind), str(p.source_id)) for p in (config.pin or ()))

        self._logical_keys_by_node_id = {
            str(node_id): frozenset((str(kind), str(source_id)) for kind, source_id in keys)
            for node_id, keys in logical_keys_by_node_id.items()
        }
        self._remaining_consumers_by_logical_key = {
            (str(kind), str(source_id)): {str(node_id) for node_id in consumers}
            for (kind, source_id), consumers in consumers_by_logical_key.items()
        }

        self._entries = OrderedDict()
        self._signature_keys_by_logical_key = {}
        self._acquired_by_node_id = {}
        self._done_node_ids = set()
        self._lock = threading.Lock()

    def get_or_load(
        self,
        signature: WorkflowCacheEntrySignature,
        *,
        workflow_node_id: str,
        load_fn: Callable[[], LoaderResultMapping],
    ) -> LoaderResultMapping:
        node_id = str(workflow_node_id)
        logical_key = signature.logical_key()
        signature_key = signature.canonical_key()
        signature_digest = signature.digest()

        conflict_diff: Optional[List[str]] = None
        conflict_target_digest: Optional[str] = None

        pending_emits: List[Tuple[str, object, Dict[str, object]]] = []

        with self._lock:
            existing_keys = self._signature_keys_by_logical_key.get(logical_key, set())
            if existing_keys and signature_key not in existing_keys:
                first_key = next(iter(existing_keys))
                first_entry = self._entries.get(first_key)
                if first_entry is not None:
                    conflict_diff = diff_signature_fields(first_entry.signature, signature)
                    conflict_target_digest = first_entry.signature.digest()

                if self._conflict_policy == "error":
                    msg = "cache_pool signature conflict for kind='{}', source_id='{}' (diff={})".format(
                        logical_key[0],
                        logical_key[1],
                        ",".join(conflict_diff or []),
                    )
                    raise ScalimWorkflowCachePoolError(msg, path="workflow.options.cache_pool.conflict_policy")

                # `warn/separate`: 允许并行存在多个条目,但需要可诊断(含差异摘要).
                warn_msg = "cache_pool signature conflict for kind='{}', source_id='{}' (policy={}, diff={})".format(
                    logical_key[0],
                    logical_key[1],
                    self._conflict_policy,
                    ",".join(conflict_diff or []),
                )
                pending_emits.append(
                    (
                        EVENT_DIAGNOSTIC_WARNING,
                        DiagnosticWarningEvent(
                            message=warn_msg,
                            source_id=str(logical_key[1]),
                            field_id=None,
                            lookup_key=None,
                            row_id=None,
                        ),
                        {
                            "workflow_exec_id": self._workflow_exec_id,
                            "workflow_node_id": node_id,
                        },
                    )
                )

            entry = self._entries.get(signature_key)
            cache_status = "hit" if entry is not None and entry.value is not None else "miss"
            if entry is None:
                self._ensure_budget_for_new_entry(workflow_node_id=node_id, pending_emits=pending_emits)
                entry = _CacheEntry(signature=signature, loading=True)
                self._entries[signature_key] = entry
                self._signature_keys_by_logical_key.setdefault(logical_key, set()).add(signature_key)

            self._acquired_by_node_id.setdefault(node_id, set()).add(signature_key)
            self._entries.move_to_end(signature_key)

            pending_emits.append(
                (
                    EVENT_WORKFLOW_CACHE_ACQUIRE,
                    WorkflowCacheAcquireEvent(
                        workflow_exec_id=self._workflow_exec_id,
                        workflow_node_id=node_id,
                        cache_kind=str(logical_key[0]),
                        source_id=str(logical_key[1]),
                        signature_digest=signature_digest,
                        cache_status=cache_status,
                        conflict_policy=str(self._conflict_policy),
                        conflict_detected=bool(conflict_diff),
                        conflict_diff_fields=tuple(conflict_diff or ()),
                        conflict_target_signature_digest=conflict_target_digest,
                    ),
                    {
                        "workflow_exec_id": self._workflow_exec_id,
                        "workflow_node_id": node_id,
                    },
                )
            )

        for event_type, payload, meta in pending_emits:
            _ = self._instrumentation.emit(str(event_type), payload, meta=meta)

        # 加载过程由每条目锁(`per-entry lock`)保护.
        with entry.lock:
            if entry.value is not None:
                return entry.value
            entry.loading = True
            try:
                loaded = load_fn()
                entry.value = loaded
                return loaded
            finally:
                entry.loading = False

    def on_workflow_node_done(self, workflow_node_id: str) -> None:
        node_id = str(workflow_node_id)

        pending_emits: List[Tuple[str, object, Dict[str, object]]] = []

        with self._lock:
            if node_id in self._done_node_ids:
                return
            self._done_node_ids.add(node_id)

            acquired = self._acquired_by_node_id.pop(node_id, set())
            pending_emits.extend(self._collect_release_events(node_id=node_id, acquired_signature_keys=acquired))
            evict_reasons = self._collect_refcount_evictions(node_id=node_id)
            for signature_key, reason in evict_reasons.items():
                pending = self._evict_entry(signature_key, workflow_node_id=node_id, reason=reason)
                if pending is not None:
                    pending_emits.append(pending)

        for event_type, payload, meta in pending_emits:
            _ = self._instrumentation.emit(str(event_type), payload, meta=meta)

    def _collect_release_events(self, *, node_id: str, acquired_signature_keys: Set[str]) -> List[Tuple[str, object, Dict[str, object]]]:
        pending_emits: List[Tuple[str, object, Dict[str, object]]] = []
        for signature_key in acquired_signature_keys:
            entry = self._entries.get(signature_key)
            if entry is None:
                continue
            logical_key = entry.signature.logical_key()
            remaining = len(self._remaining_consumers_by_logical_key.get(logical_key, set()))
            pending_emits.append(
                (
                    EVENT_WORKFLOW_CACHE_RELEASE,
                    WorkflowCacheReleaseEvent(
                        workflow_exec_id=self._workflow_exec_id,
                        workflow_node_id=node_id,
                        cache_kind=str(logical_key[0]),
                        source_id=str(logical_key[1]),
                        signature_digest=entry.signature.digest(),
                        remaining_consumers=remaining,
                        release_policy=str(self._release_policy),
                        is_pinned=logical_key in self._pinned_logical_keys,
                    ),
                    {
                        "workflow_exec_id": self._workflow_exec_id,
                        "workflow_node_id": node_id,
                    },
                )
            )
        return pending_emits

    def _collect_refcount_evictions(self, *, node_id: str) -> Dict[str, str]:
        evict_reasons: Dict[str, str] = {}
        for logical_key in self._logical_keys_by_node_id.get(node_id, frozenset()):
            remaining = self._remaining_consumers_by_logical_key.get(logical_key)
            if remaining is None:
                continue
            remaining.discard(node_id)
            if remaining:
                continue
            if self._release_policy != "dag_refcount":
                continue
            if logical_key in self._pinned_logical_keys:
                continue
            for signature_key in list(self._signature_keys_by_logical_key.get(logical_key, set())):
                entry = self._entries.get(signature_key)
                if entry is not None and entry.loading:
                    continue
                evict_reasons[signature_key] = "refcount_zero"
        return evict_reasons

    def close(self) -> None:
        pending_emits: List[Tuple[str, object, Dict[str, object]]] = []
        with self._lock:
            for signature_key in list(self._entries.keys()):
                pending = self._evict_entry(signature_key, workflow_node_id="workflow_end", reason="workflow_end")
                if pending is not None:
                    pending_emits.append(pending)

        for event_type, payload, meta in pending_emits:
            _ = self._instrumentation.emit(str(event_type), payload, meta=meta)

    def _ensure_budget_for_new_entry(self, *, workflow_node_id: str, pending_emits: List[Tuple[str, object, Dict[str, object]]]) -> None:
        if self._max_entries < 1:  # pragma: no cover  # pragma: allow-no-cover invariant: budget validated by config loader
            msg = "cache_pool budget.max_entries must be >= 1"
            raise ScalimWorkflowCachePoolError(msg, path="workflow.options.cache_pool.budget.max_entries")
        if len(self._entries) < self._max_entries:
            return

        if self._over_budget_policy == "fail_fast":
            msg = "cache_pool over budget: max_entries={} (over_budget_policy=fail_fast)".format(self._max_entries)
            raise ScalimWorkflowCachePoolError(msg, path="workflow.options.cache_pool.budget.over_budget_policy")

        if self._over_budget_policy != "evict_lru":
            msg = "cache_pool over_budget_policy '{}' is not supported".format(self._over_budget_policy)
            raise ScalimWorkflowCachePoolError(msg, path="workflow.options.cache_pool.budget.over_budget_policy")

        evicted = self._evict_lru_idle(workflow_node_id=workflow_node_id, pending_emits=pending_emits)
        if not evicted:
            msg = "cache_pool over budget: max_entries={} (no evictable refcount=0 entries)".format(self._max_entries)
            raise ScalimWorkflowCachePoolError(msg, path="workflow.options.cache_pool.budget.over_budget_policy")

    def _evict_lru_idle(self, *, workflow_node_id: str, pending_emits: List[Tuple[str, object, Dict[str, object]]]) -> bool:
        for signature_key, entry in list(self._entries.items()):
            logical_key = entry.signature.logical_key()
            if logical_key in self._pinned_logical_keys:
                continue
            if entry.loading:
                continue
            remaining = self._remaining_consumers_by_logical_key.get(logical_key, set())
            if remaining:
                continue
            pending = self._evict_entry(signature_key, workflow_node_id=workflow_node_id, reason="budget_lru")
            if pending is not None:
                pending_emits.append(pending)
            return True
        return False

    def _evict_entry(
        self,
        signature_key: str,
        *,
        workflow_node_id: str,
        reason: str,
    ) -> Optional[Tuple[str, object, Dict[str, object]]]:
        entry = self._entries.pop(signature_key, None)
        if entry is None:
            return None
        logical_key = entry.signature.logical_key()
        keys = self._signature_keys_by_logical_key.get(logical_key)
        if keys is not None:
            keys.discard(signature_key)
            if not keys:
                _ = self._signature_keys_by_logical_key.pop(logical_key, None)
                _ = self._remaining_consumers_by_logical_key.pop(logical_key, None)

        return (
            EVENT_WORKFLOW_CACHE_EVICT,
            WorkflowCacheEvictEvent(
                workflow_exec_id=self._workflow_exec_id,
                workflow_node_id=str(workflow_node_id),
                cache_kind=str(logical_key[0]),
                source_id=str(logical_key[1]),
                signature_digest=entry.signature.digest(),
                reason=str(reason),
            ),
            {
                "workflow_exec_id": self._workflow_exec_id,
                "workflow_node_id": str(workflow_node_id),
            },
        )


__all__ = [
    "ScalimWorkflowCachePoolError",
    "WorkflowCacheEntrySignature",
    "WorkflowCachePool",
    "build_preload_forever_signature",
    "diff_signature_fields",
]
